//
//  NSMutableURLRequest+SyncKitAdditions.m
//  SyncKit
//
//  Created by Logan Collins on 4/9/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "NSMutableURLRequest+SyncKitAdditions.h"

#import "SNKCredentials_Private.h"

#import "NSData+SyncKitAdditions.h"
#import "NSDictionary+SyncKitAdditions.h"
#import "NSString+SyncKitAdditions.h"
#import "NSURL+SyncKitAdditions.h"


@implementation NSMutableURLRequest (SyncKitAdditions)

- (NSDictionary *)snk_parameters {
    NSString *encodedParameters = nil;
    
    if ([[self HTTPMethod] isEqualToString:@"GET"] || [[self HTTPMethod] isEqualToString:@"DELETE"]) {
        // GET, DELETE
        encodedParameters = [[self URL] query];
    }
    else if ([[self HTTPMethod] isEqualToString:@"POST"] || [[self HTTPMethod] isEqualToString:@"PUT"]) {
        // POST, PUT
        if ([self valueForHTTPHeaderField:@"Content-Type"] == nil
            || [[self valueForHTTPHeaderField:@"Content-Type"] isEqualToString:@"application/x-www-form-urlencoded"]) {
            encodedParameters = [[NSString alloc] initWithData:[self HTTPBody] encoding:NSUTF8StringEncoding];
        }
        else {
            encodedParameters = [[self URL] query];
        }
    }
    
    if ((encodedParameters == nil) || ([encodedParameters isEqualToString:@""])) {
        return nil;
    }
    
    NSArray *encodedParameterPairs = [encodedParameters componentsSeparatedByString:@"&"];
    NSMutableDictionary *requestParameters = [NSMutableDictionary dictionary];
    
    for (NSString *encodedPair in encodedParameterPairs) {
        NSArray *encodedPairElements = [encodedPair componentsSeparatedByString:@"="];
        NSString *key = [encodedPairElements objectAtIndex:0];
        NSString *value = [encodedPairElements objectAtIndex:1];
        [requestParameters setObject:value forKey:key];
    }
    
    return requestParameters;
}

- (void)snk_signWithOAuthConsumerKey:(NSString *)consumerKey consumerSecret:(NSString *)consumerSecret credentials:(SNKCredentials *)credentials {
    if (credentials != nil && credentials.type != SNKCredentialsTypeOAuth) {
        @throw [NSException exceptionWithName:NSInvalidArgumentException reason:@"Invalid credentials" userInfo:nil];
    }
    
    // Credentials
    NSString *token = [credentials OAuthToken];
    if (token == nil) {
        token = @"";
    }
    
    NSString *tokenSecret = [credentials OAuthTokenSecret];
    if (tokenSecret == nil) {
        tokenSecret = @"";
    }
    
    NSString *secret = [NSString stringWithFormat:@"%@&%@", [consumerSecret snk_URLEncodedString], [tokenSecret snk_URLEncodedString]];
    NSData *secretData = [secret dataUsingEncoding:NSUTF8StringEncoding];
    
    NSString *timestamp = [NSString stringWithFormat:@"%ld", (long)time(NULL)];
    CFUUIDRef UUID = CFUUIDCreate(NULL);
    NSString *UUIDString = CFBridgingRelease(CFUUIDCreateString(NULL, UUID));
    CFRelease(UUID);
    
    
    // Signature
    
    // OAuth Spec, Section 9.1.1 "Normalize Request Parameters"
    // Build a sorted array of both request parameters and OAuth header parameters
    NSMutableArray *parameterPairs = [NSMutableArray array];
    
    [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_consumer_key", consumerKey]];
    [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_signature_method", @"HMAC-SHA1"]];
    [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_timestamp", timestamp]];
    [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_nonce", UUIDString]];
    [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_version", @"1.0"]];
    if ([token length] > 0) {
        [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", @"oauth_token", token]];
    }
    
    NSDictionary *parameters = [self snk_parameters];
    for (NSString *key in parameters) {
        NSString *value = parameters[key];
        [parameterPairs addObject:[NSString stringWithFormat:@"%@=%@", key, value]];
    }
    
    NSString *normalizedRequestParameters = [[parameterPairs sortedArrayUsingSelector:@selector(caseInsensitiveCompare:)] componentsJoinedByString:@"&"];
    
    // OAuth Spec, Section 9.1.2 "Concatenate Request Elements"
    NSString *signatureBaseString = [NSString stringWithFormat:@"%@&%@&%@", [self HTTPMethod], [[[self URL] snk_URLStringWithoutQuery] snk_URLEncodedString], [normalizedRequestParameters snk_URLEncodedString]];
    NSData *signatureData = [[signatureBaseString dataUsingEncoding:NSUTF8StringEncoding] snk_HMAC_SHA1WithKey:secretData];
    NSString *signature = [signatureData snk_base64String];
    
    
    // OAuth parameters
    NSString *oauthToken = nil;
    if ([token length] == 0) {
        oauthToken = @"";
    }
    else {
        oauthToken = [NSString stringWithFormat:@"oauth_token=\"%@\", ", [token snk_URLEncodedString]];
    }
    
    NSString *oauthHeader = [NSString stringWithFormat:@"OAuth oauth_consumer_key=\"%@\", %@oauth_signature_method=\"HMAC-SHA1\", oauth_signature=\"%@\", oauth_timestamp=\"%@\", oauth_nonce=\"%@\", oauth_version=\"1.0\"",
                             [consumerKey snk_URLEncodedString],
                             oauthToken,
                             [signature snk_URLEncodedString],
                             [timestamp snk_URLEncodedString],
                             [UUIDString snk_URLEncodedString]];
    
    [self setValue:oauthHeader forHTTPHeaderField:@"Authorization"];
}

@end
