//
//  SNKChangeContext.m
//  SyncKit
//
//  Created by Logan Collins on 10/26/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "SNKChangeContext.h"
#import "SNKChangeContext_Private.h"

#import "SNKChangeCoordinator_Private.h"

#import "SNKChangeObject_Private.h"
#import "SNKChange_Private.h"
#import "SNKChangeEntity_Private.h"


@implementation SNKChangeContext {
    NSManagedObjectContext *_managedObjectContext;
}

- (id)initWithChangeCoordinator:(SNKChangeCoordinator *)coordinator {
    NSParameterAssert(coordinator != nil);
    
    self = [super init];
    if (self) {
        self.changeCoordinator = coordinator;
        
        _managedObjectContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
        _managedObjectContext.persistentStoreCoordinator = coordinator.persistentStoreCoordinator;
        _managedObjectContext.mergePolicy = [[NSMergePolicy alloc] initWithMergeType:NSMergeByPropertyObjectTrumpMergePolicyType];
    }
    return self;
}


#pragma mark -
#pragma mark Operations

- (void)performBlock:(void (^)(void))block {
    [_managedObjectContext performBlock:block];
}

- (void)performBlockAndWait:(void (^)(void))block {
    [_managedObjectContext performBlockAndWait:block];
}


#pragma mark -
#pragma mark Persistence

- (BOOL)commit:(NSError *__autoreleasing *)outError {
    return [_managedObjectContext save:outError];
}

- (void)rollback {
    [_managedObjectContext rollback];
}


#pragma mark -
#pragma mark Entities

- (NSArray *)entities {
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Entity"];
    NSArray *results = [_managedObjectContext executeFetchRequest:request error:nil];
    
    NSMutableArray *entities = [NSMutableArray arrayWithCapacity:[results count]];
    for (NSManagedObject *managedObject in results) {
        SNKChangeEntity *entity = [SNKChangeEntity objectWithManagedObject:managedObject context:self];
        if (entity != nil) {
            [entities addObject:entity];
        }
    }
    return entities;
}

- (SNKChangeEntity *)entityWithName:(NSString *)entityName create:(BOOL)create {
    NSParameterAssert(entityName != nil && [entityName length] > 0);
    
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Entity"];
    request.fetchLimit = 1;
    request.predicate = [NSPredicate predicateWithFormat:@"name == %@", entityName];
    
    NSArray *results = [_managedObjectContext executeFetchRequest:request error:nil];
    NSManagedObject *managedObject = [results lastObject];
    if (managedObject == nil) {
        managedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Entity" inManagedObjectContext:_managedObjectContext];
        [managedObject setValue:entityName forKey:@"name"];
    }
    return [SNKChangeEntity objectWithManagedObject:managedObject context:self];
}

- (void)deleteEntity:(SNKChangeEntity *)entity {
    [_managedObjectContext deleteObject:entity.managedObject];
}


#pragma mark -
#pragma mark Changes

- (NSPredicate *)managedObjectPredicateForPredicate:(NSPredicate *)predicate {
    if ([predicate isKindOfClass:[NSCompoundPredicate class]]) {
        NSArray *subpredicates = [(NSCompoundPredicate *)predicate subpredicates];
        NSMutableArray *newSubpredicates = [NSMutableArray arrayWithCapacity:[subpredicates count]];
        for (NSPredicate *subpredicate in subpredicates) {
            NSPredicate *newSubpredicate = [self managedObjectPredicateForPredicate:subpredicate];
            [newSubpredicates addObject:newSubpredicate];
        }
        return [[NSCompoundPredicate alloc] initWithType:[(NSCompoundPredicate *)predicate compoundPredicateType] subpredicates:newSubpredicates];
    }
    else if ([predicate isKindOfClass:[NSComparisonPredicate class]]) {
        NSPredicate *newPredicate = nil;
        NSExpression *newRightExpression = nil;
        
        NSExpression *rightExpression = [(NSComparisonPredicate *)predicate rightExpression];
        if ([rightExpression expressionType] == NSConstantValueExpressionType) {
            id value = [rightExpression constantValue];
            if ([value isKindOfClass:[SNKChangeObject class]]) {
                newRightExpression = [NSExpression expressionForConstantValue:[value managedObject]];
            }
            else {
                newRightExpression = rightExpression;
            }
        }
        else {
            newRightExpression = rightExpression;
        }
        
        if (newRightExpression != rightExpression) {
            newPredicate = [[NSComparisonPredicate alloc] initWithLeftExpression:[(NSComparisonPredicate *)predicate leftExpression]
                                                                 rightExpression:newRightExpression
                                                                        modifier:[(NSComparisonPredicate *)predicate comparisonPredicateModifier]
                                                                            type:[(NSComparisonPredicate *)predicate predicateOperatorType]
                                                                         options:[(NSComparisonPredicate *)predicate options]];
        }
        else {
            newPredicate = predicate;
        }
        
        return newPredicate;
    }
    else {
        return predicate;
    }
}

- (NSArray *)changesMatchingPredicate:(NSPredicate *)predicate {
    predicate = [self managedObjectPredicateForPredicate:predicate];
    
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Change"];
    request.predicate = predicate;
    request.sortDescriptors = @[[NSSortDescriptor sortDescriptorWithKey:@"date" ascending:YES]];
    NSArray *results = [_managedObjectContext executeFetchRequest:request error:nil];
    NSMutableArray *changes = [NSMutableArray arrayWithCapacity:[results count]];
    for (NSManagedObject *managedObject in results) {
        SNKChange *change = [SNKChange objectWithManagedObject:managedObject context:self];
        if (change != nil) {
            [changes addObject:change];
        }
    }
    return changes;
}

- (void)insertedRecordWithIdentifier:(NSString *)recordIdentifier entity:(NSString *)entityName attributes:(NSSet *)attributes {
    NSParameterAssert(recordIdentifier != nil && [recordIdentifier length] > 0);
    NSParameterAssert(entityName != nil && [entityName length] > 0);
    
    SNKChangeEntity *entity = [self entityWithName:entityName create:YES];
    
    NSManagedObject *managedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Change" inManagedObjectContext:_managedObjectContext];
    [managedObject setValue:entity.managedObject forKey:@"changeEntity"];
    [managedObject setValue:@(SNKChangeTypeInsert) forKey:@"type"];
    [managedObject setValue:recordIdentifier forKey:@"identifier"];
    [managedObject setValue:[NSDate date] forKey:@"date"];
    [managedObject setValue:attributes forKey:@"attributes"];
}

- (void)updatedRecordWithIdentifier:(NSString *)recordIdentifier entity:(NSString *)entityName attributes:(NSSet *)attributes {
    NSParameterAssert(recordIdentifier != nil && [recordIdentifier length] > 0);
    NSParameterAssert(entityName != nil && [entityName length] > 0);
    
    SNKChangeEntity *entity = [self entityWithName:entityName create:YES];
    
    NSManagedObject *managedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Change" inManagedObjectContext:_managedObjectContext];
    [managedObject setValue:entity.managedObject forKey:@"changeEntity"];
    [managedObject setValue:@(SNKChangeTypeUpdate) forKey:@"type"];
    [managedObject setValue:recordIdentifier forKey:@"identifier"];
    [managedObject setValue:[NSDate date] forKey:@"date"];
    [managedObject setValue:attributes forKey:@"attributes"];
}

- (void)deletedRecordWithIdentifier:(NSString *)recordIdentifier entity:(NSString *)entityName {
    NSParameterAssert(recordIdentifier != nil && [recordIdentifier length] > 0);
    NSParameterAssert(entityName != nil && [entityName length] > 0);
    
    SNKChangeEntity *entity = [self entityWithName:entityName create:YES];
    
    NSManagedObject *managedObject = [NSEntityDescription insertNewObjectForEntityForName:@"Change" inManagedObjectContext:_managedObjectContext];
    [managedObject setValue:entity.managedObject forKey:@"changeEntity"];
    [managedObject setValue:@(SNKChangeTypeDelete) forKey:@"type"];
    [managedObject setValue:recordIdentifier forKey:@"identifier"];
    [managedObject setValue:[NSDate date] forKey:@"date"];
}

- (void)deleteChange:(SNKChange *)change {
    [_managedObjectContext deleteObject:change.managedObject];
}

- (void)deleteAllChangesForEntity:(SNKChangeEntity *)entity {
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Change"];
    request.predicate = [NSPredicate predicateWithFormat:@"entity == %@", entity.managedObject];
    NSArray *results = [_managedObjectContext executeFetchRequest:request error:nil];
    for (NSManagedObject *managedObject in results) {
        [_managedObjectContext deleteObject:managedObject];
    }
}

- (void)deleteAllChanges {
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Change"];
    NSArray *results = [_managedObjectContext executeFetchRequest:request error:nil];
    for (NSManagedObject *managedObject in results) {
        [_managedObjectContext deleteObject:managedObject];
    }
}

@end
