//
//  SNKChangeObject.m
//  SyncKit
//
//  Created by Logan Collins on 10/26/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "SNKChangeObject.h"
#import "SNKChangeObject_Private.h"

#import "SNKChangeContext_Private.h"

#import <objc/runtime.h>


static NSString * SNKObjectManagedPropertyObservationContext = @"SNKObjectManagedPropertyObservationContext";


static id id_id_sel_tramp(id _s, SEL _c, ...) {
	id (^block)(id) = objc_getAssociatedObject([_s class], _c);
	if (block != NULL) {
		return block(_s);
	}
	return nil;
}

static id void_id_sel_id_tramp(id _s, SEL _c, ...) {
	va_list ap;
	va_start(ap, _c);
	id v = va_arg(ap, id);
	va_end(ap);
	
	void (^block)(id, id) = objc_getAssociatedObject([_s class], _c);
	if (block != NULL) {
		block(_s, v);
	}
	
	return nil;
}


@implementation SNKChangeObject

@synthesize context=_context;
@synthesize managedObject=_managedObject;

+ (void)initializeWithEntity:(NSEntityDescription *)entity {
	NSArray *properties = [entity properties];
	for (NSPropertyDescription *property in properties) {
		NSString *propertyName = [property name];
		
		// Add dynamic method implementations for attributes and relationships
		if ([property isKindOfClass:[NSAttributeDescription class]]
			|| [property isKindOfClass:[NSRelationshipDescription class]]) {
			
			// Synthesize getter
			SEL getterSelector = NSSelectorFromString(propertyName);
			if (class_getInstanceMethod(self, getterSelector) == NULL) {
				id (^implementation)(id) = ^id(id _self) {
					return [_self managedValueForKey:propertyName];
				};
				objc_setAssociatedObject(self, getterSelector, [implementation copy], OBJC_ASSOCIATION_RETAIN);
				class_addMethod(self, getterSelector, id_id_sel_tramp, "@@:");
			}
			
			// Synthesize setter
			SEL setterSelector = NSSelectorFromString([NSString stringWithFormat:@"set%@%@:", [[propertyName substringToIndex:1] uppercaseString], [propertyName substringFromIndex:1]]);
			if (class_getInstanceMethod(self, setterSelector) == NULL) {
				void (^implementation)(id, id) = ^(id _self, id value) {
					[_self setManagedValue:value forKey:propertyName];
				};
				objc_setAssociatedObject(self, setterSelector, [implementation copy], OBJC_ASSOCIATION_RETAIN);
				class_addMethod(self, setterSelector, void_id_sel_id_tramp, "v@:@");
			}
		}
	}
}

+ (instancetype)objectWithManagedObject:(NSManagedObject *)managedObject context:(SNKChangeContext *)context {
	SNKChangeObject *object = [managedObject snk_changeObject];
	if (object == nil) {
		Class objectClass = [[managedObject entity] snk_changeObjectClass];
		object = [[objectClass alloc] initWithManagedObject:managedObject context:context];
	}
	return object;
}

- (id)initWithManagedObject:(NSManagedObject *)managedObject context:(SNKChangeContext *)context {
	NSParameterAssert(managedObject != nil);
	NSParameterAssert(context != nil);
	
	self = [super init];
	if (self) {
		_attributes = [[NSMutableDictionary alloc] init];
		
		self.context = context;
		
		self.managedObject = managedObject;
		[managedObject snk_setChangeObject:self];
		
		// Add observations
		NSEntityDescription *entity = [managedObject entity];
		for (NSString *propertyName in [entity propertiesByName]) {
			NSPropertyDescription *property = [[entity propertiesByName] objectForKey:propertyName];
			if ([property isKindOfClass:[NSAttributeDescription class]]) {
				// Prefetch attributes
				id value = [managedObject valueForKey:propertyName];
				if (value != nil) {
					[_attributes setObject:value forKey:propertyName];
				}
				else {
					[_attributes setObject:[NSNull null] forKey:propertyName];
				}
			}
			
			[managedObject addObserver:self forKeyPath:propertyName options:(NSKeyValueObservingOptionNew) context:(__bridge void *)SNKObjectManagedPropertyObservationContext];
		}
	}
	return self;
}

- (id)init {
	@throw [NSException exceptionWithName:NSInternalInconsistencyException reason:@"SNKChangeObject instances cannot be initialized directly." userInfo:nil];
}

- (void)dealloc {
	// Remove observations
	NSEntityDescription *entity = [_managedObject entity];
	for (NSString *propertyName in [entity propertiesByName]) {
		[_managedObject removeObserver:self forKeyPath:propertyName context:(__bridge void *)SNKObjectManagedPropertyObservationContext];
	}
	
	[_managedObject snk_setChangeObject:nil];
	
#if !__has_feature(objc_arc)
	[_managedObject release];
	[_attributes release];
	[super dealloc];
#endif
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context {
    if (context == (__bridge void *)SNKObjectManagedPropertyObservationContext) {
		id value = change[NSKeyValueChangeNewKey];
		
		[self willChangeValueForKey:keyPath];
		
		// Autocache
		NSEntityDescription *entity = [_managedObject entity];
		NSPropertyDescription *property = [[entity propertiesByName] objectForKey:keyPath];
		if ([property isKindOfClass:[NSAttributeDescription class]]) {
			if (value != nil) {
				[_attributes setObject:value forKey:keyPath];
			}
			else {
				[_attributes setObject:[NSNull null] forKey:keyPath];
			}
		}
		else {
			[_attributes removeObjectForKey:keyPath];
		}
		
		[self didChangeValueForKey:keyPath];
    }
	else {
        [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
}

- (BOOL)hasChanges {
	return ([_managedObject isInserted] || [_managedObject isUpdated] || [_managedObject isDeleted]);
}

- (BOOL)isNew {
	return [_managedObject isInserted];
}

- (BOOL)refresh {
	[self.context.managedObjectContext refreshObject:self.managedObject mergeChanges:YES];
	[_attributes removeAllObjects];
	return (![_managedObject isDeleted]);
}

- (void)reset {
	[self.context.managedObjectContext refreshObject:self.managedObject mergeChanges:NO];
	[_attributes removeAllObjects];
}

- (void)rollback {
	[_attributes removeAllObjects];
}


#pragma mark -
#pragma mark Key-Value Coding

+ (BOOL)accessInstanceVariablesDirectly {
	return NO;
}

- (id)valueForKey:(NSString *)key {
	return [self managedValueForKey:key];
}

- (void)setValue:(id)value forKey:(NSString *)key {
	[self setManagedValue:value forKey:key];
}

- (id)managedValueForKey:(NSString *)key {
	id value = [_attributes objectForKey:key];
	
	if ([value isKindOfClass:[NSNull class]]) {
		// <null>
		value = nil;
	}
	else if (value == nil) {
		NSEntityDescription *entity = [_managedObject entity];
		if ([[entity attributesByName] objectForKey:key] != nil) {
			// Attribute
			value = [_managedObject valueForKey:key];
		}
		else if ([[entity relationshipsByName] objectForKey:key] != nil) {
			// Relationship
			NSRelationshipDescription *relationship = [[entity relationshipsByName] objectForKey:key];
			if ([relationship isToMany]) {
				// To-many
				if ([relationship isOrdered]) {
					// Ordered
					NSOrderedSet *orderedSet = [_managedObject valueForKey:key];
					NSMutableOrderedSet *newOrderedSet = [NSMutableOrderedSet orderedSetWithCapacity:[orderedSet count]];
					for (NSManagedObject *managedObject in orderedSet) {
						SNKChangeObject *object = [SNKChangeObject objectWithManagedObject:managedObject context:self.context];
						[newOrderedSet addObject:object];
					}
					value = newOrderedSet;
				}
				else {
					// Unordered
					NSSet *set = [_managedObject valueForKey:key];
					NSMutableOrderedSet *newSet = [NSMutableOrderedSet orderedSetWithCapacity:[set count]];
					for (NSManagedObject *managedObject in set) {
						SNKChangeObject *object = [SNKChangeObject objectWithManagedObject:managedObject context:self.context];
						[newSet addObject:object];
					}
					value = newSet;
				}
			}
			else {
				// To-one
				NSManagedObject *managedObject = [_managedObject valueForKey:key];
				if (managedObject != nil) {
					value = [SNKChangeObject objectWithManagedObject:managedObject context:self.context];
				}
			}
		}
		
		if (value != nil) {
			// Cache value
			[_attributes setObject:value forKey:key];
		}
		else {
			// Cache <null> to prevent refetches
			[_attributes setObject:[NSNull null] forKey:key];
		}
	}
	
	return value;
}

- (void)setManagedValue:(id)value forKey:(NSString *)key {
	if (value == nil) {
		[_attributes setObject:[NSNull null] forKey:key];
		[_managedObject setValue:nil forKey:key];
	}
	else {
		if (value != nil) {
			// Cache value
			[_attributes setObject:value forKey:key];
		}
		else {
			// Cache <null> to prevent refetches
			[_attributes setObject:[NSNull null] forKey:key];
		}
		
		NSEntityDescription *entity = [_managedObject entity];
		if ([[entity attributesByName] objectForKey:key] != nil) {
			// Attribute
			[_managedObject setValue:value forKey:key];
		}
		else if ([[entity relationshipsByName] objectForKey:key] != nil) {
			// Relationship
			NSRelationshipDescription *relationship = [[entity relationshipsByName] objectForKey:key];
			if ([relationship isToMany]) {
				// To-many
				if ([relationship isOrdered]) {
					// Ordered
					NSOrderedSet *orderedSet = value;
					NSMutableOrderedSet *newOrderedSet = [NSMutableOrderedSet orderedSetWithCapacity:[orderedSet count]];
					for (SNKChangeObject *object in orderedSet) {
						NSManagedObject *managedObject = [object managedObject];
						[newOrderedSet addObject:managedObject];
					}
					[_managedObject setValue:newOrderedSet forKey:key];
				}
				else {
					// Unordered
					NSSet *set = value;
					NSMutableSet *newSet = [NSMutableSet setWithCapacity:[set count]];
					for (SNKChangeObject *object in set) {
						NSManagedObject *managedObject = [object managedObject];
						[newSet addObject:managedObject];
					}
					[_managedObject setValue:newSet forKey:key];
				}
			}
			else {
				// To-one
				NSManagedObject *managedObject = [value managedObject];
				[_managedObject setValue:managedObject forKey:key];
			}
		}
	}
}


#pragma mark -
#pragma mark Exporting

- (NSDictionary *)serializableDictionaryRepresentation {
	// Overridden by subclasses
	return @{};
}

- (void)applySerializableDictionaryRepresentation:(NSDictionary *)dictionary {
	// Overridden by subclasses
}

@end


@implementation NSManagedObject (SNKChangeObjectAdditions)

static NSString * const SNKChangeObjectKey = @"SNKChangeObject";

- (SNKChangeObject *)snk_changeObject {
	return objc_getAssociatedObject(self, (__bridge void *)SNKChangeObjectKey);
}

- (void)snk_setChangeObject:(SNKChangeObject *)object {
	objc_setAssociatedObject(self, (__bridge void *)SNKChangeObjectKey, object, OBJC_ASSOCIATION_ASSIGN);
}

@end


@implementation NSEntityDescription (SNKChangeObjectAdditions)

static NSString * const SNKChangeObjectClassKey = @"com.sunflowersw.synckit.change-object-class";

- (Class)snk_recursiveChangeObjectClass {
    Class result = Nil;
    id value = [[self userInfo] objectForKey:SNKChangeObjectClassKey];
    if ([value isKindOfClass:[NSString class]]) {
        result = NSClassFromString(value);
    }
    
    if (result == Nil) {
        result = [[self superentity] snk_recursiveChangeObjectClass];
    }
    
    return result;
}

- (Class)snk_changeObjectClass {
    Class objectClass = objc_getAssociatedObject(self, (__bridge void *)SNKChangeObjectClassKey);
    if (objectClass == Nil) {
        objectClass = [self snk_recursiveChangeObjectClass];
        if (objectClass == Nil) {
            objectClass = [[self superentity] snk_changeObjectClass];
        }
		if (objectClass == Nil) {
			objectClass = [SNKChangeObject class];
		}
		[objectClass initializeWithEntity:self];
        objc_setAssociatedObject(self, (__bridge void *)SNKChangeObjectClassKey, objectClass, OBJC_ASSOCIATION_COPY);
    }
    return objectClass;
}

@end
