//
//  SNKClient.m
//  SyncKit
//
//  Created by Logan Collins on 4/5/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "SNKClient.h"
#import "SNKClient_Private.h"

#if TARGET_OS_IPHONE
#import <UIKit/UIKit.h>
#import <sys/utsname.h>
#else
#import <IOKit/IOKitLib.h>
#import <sys/sysctl.h>
#endif


@implementation SNKClient

- (id)initWithDictionary:(NSDictionary *)dictionary {
    self = [super init];
    if (self) {
        self.identifier = ([dictionary[@"identifier"] isKindOfClass:[NSString class]] ? dictionary[@"identifier"] : nil);
        self.name = ([dictionary[@"name"] isKindOfClass:[NSString class]] ? dictionary[@"name"] : nil);
        self.model = ([dictionary[@"model"] isKindOfClass:[NSString class]] ? dictionary[@"model"] : nil);
        
        id systemType = dictionary[@"system_type"];
        if ([systemType isKindOfClass:[NSString class]]) {
            if ([systemType isEqualToString:@"ios"]) {
                self.type = SNKClientTypeiOS;
            }
            else if ([systemType isEqualToString:@"mac"]) {
                self.type = SNKClientTypeMac;
            }
        }
        
        self.systemVersion = ([dictionary[@"system_version"] isKindOfClass:[NSString class]] ? dictionary[@"system_version"] : nil);
        self.localeIdentifier = ([dictionary[@"locale"] isKindOfClass:[NSString class]] ? dictionary[@"locale"] : nil);
        
        id applicationInfo = dictionary[@"app"];
        if ([applicationInfo isKindOfClass:[NSDictionary class]]) {
            self.applicationIdentifier = ([applicationInfo[@"identifier"] isKindOfClass:[NSString class]] ? applicationInfo[@"identifier"] : nil);
            self.applicationName = ([applicationInfo[@"name"] isKindOfClass:[NSString class]] ? applicationInfo[@"name"] : nil);
            self.applicationVersion = ([applicationInfo[@"version"] isKindOfClass:[NSString class]] ? applicationInfo[@"version"] : nil);
        }
        
        self.iconsDictionary = dictionary[@"icons"];
        
        static NSDateFormatter *dateFormatter = nil;
        if (dateFormatter == nil) {
            dateFormatter = [[NSDateFormatter alloc] init];
            [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
            
            NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
            [dateFormatter setLocale:locale];
        }
        
        id lastSyncValue = dictionary[@"last_sync"];
        if ([lastSyncValue isKindOfClass:[NSString class]]) {
            self.lastSyncDate = [dateFormatter dateFromString:lastSyncValue];
        }
    }
    return self;
}

- (NSString *)description {
    return [NSString stringWithFormat:@"<%@: %p : identifier=%@, name=%@, lastSyncDate=%@>", [self class], self, self.identifier, self.name, self.lastSyncDate];
}

@end


NSString * SNKLocalClientIdentifier() {
    NSString *identifier = nil;
#if TARGET_OS_IPHONE
//    if ([UIDevice instancesRespondToSelector:@selector(identifierForVendor)]) {
        identifier = [[[UIDevice currentDevice] identifierForVendor] UUIDString];
//    }
//    else {
//#pragma clang diagnostic push
//#pragma clang diagnostic ignored "-Wdeprecated-declarations"
//        identifier = [[UIDevice currentDevice] uniqueIdentifier];
//#pragma clang diagnostic pop
//    }
#else
    // Get the unique platform identifier from IOKit
    io_service_t platformExpert = IOServiceGetMatchingService(kIOMasterPortDefault, IOServiceMatching("IOPlatformExpertDevice"));
    if (platformExpert != IO_OBJECT_NULL) {
        identifier = CFBridgingRelease((CFStringRef)IORegistryEntryCreateCFProperty(platformExpert, CFSTR(kIOPlatformUUIDKey), kCFAllocatorDefault, 0));
        IOObjectRelease(platformExpert);
    }
#endif
    return identifier;
}


NSDictionary * SNKLocalClientDictionary() {
    NSMutableDictionary *dictionary = [NSMutableDictionary dictionary];
    
#if TARGET_OS_IPHONE
    struct utsname systemInfo;
    uname(&systemInfo);
    NSString *model = [NSString stringWithCString:systemInfo.machine encoding:NSUTF8StringEncoding];
    if (model != nil) {
        dictionary[@"model"] = model;
    }
    
    NSString *deviceName = [[UIDevice currentDevice] name];
    if (deviceName != nil) {
        dictionary[@"name"] = deviceName;
    }
    
    dictionary[@"system_type"] = @"ios";
    dictionary[@"system_version"] = [[UIDevice currentDevice] systemVersion];
#else
    // Get the hardware model identifier from sysctl
    size_t len = 0;
    sysctlbyname("hw.model", NULL, &len, NULL, 0);
    if (len > 0) {
        char * model = malloc(len * sizeof(char));
        sysctlbyname("hw.model", model, &len, NULL, 0);
        dictionary[@"model"] = [NSString stringWithCString:model encoding:NSUTF8StringEncoding];
        free(model);
    }
    
    NSString *deviceName = [[NSHost currentHost] localizedName];
    if (deviceName != nil) {
        dictionary[@"name"] = deviceName;
    }
    
    dictionary[@"system_type"] = @"mac";
    
    // Get the system version from CoreServices
    NSURL *systemVersionURL = [[[NSFileManager defaultManager] URLForDirectory:NSLibraryDirectory inDomain:NSSystemDomainMask appropriateForURL:nil create:NO error:nil] URLByAppendingPathComponent:@"CoreServices/SystemVersion.plist"];
    NSData *systemVersionData = [NSData dataWithContentsOfURL:systemVersionURL];
    NSDictionary *systemInfoDict = [NSPropertyListSerialization propertyListWithData:systemVersionData options:0 format:NULL error:nil];
    if (systemInfoDict != nil) {
        NSString *productVersion = systemInfoDict[@"ProductVersion"];
        if ([productVersion length] > 0) {
            dictionary[@"system_version"] = productVersion;
        }
    }
#endif
    
    dictionary[@"locale"] = [[NSLocale currentLocale] localeIdentifier];
    
    NSBundle *bundle = [NSBundle mainBundle];
    NSMutableDictionary *application = [NSMutableDictionary dictionary];
    application[@"identifier"] = [bundle objectForInfoDictionaryKey:@"CFBundleIdentifier"];
    application[@"name"] = [bundle objectForInfoDictionaryKey:@"CFBundleName"];
    application[@"version"] = [bundle objectForInfoDictionaryKey:@"CFBundleShortVersionString"];
    dictionary[@"application"] = application;
    
    return dictionary;
}
