//
//  SNKPersistentSyncCoordinator.h
//  SyncKit
//
//  Created by Logan Collins on 3/26/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <CoreData/CoreData.h>
#import <SyncKit/SyncKitDefines.h>


@class SNKService, SNKSyncSession, SNKRecord, SNKProperty;
@protocol SNKPersistentSyncCoordinatorDelegate;


/*!
 * @class SNKPersistentSyncCoordinator
 * @abstract Automatically coordinates sync for Core Data stores
 * 
 * @discussion
 * A persistent sync coordinator works in tandem with a primary managed
 * object context to monitor for changes in the object graph, relaying these
 * changes to a sync service, and applying changes received from the sync
 * service automatically.
 * 
 * To support syncing, a persistent sync coordinator must be created for 
 * a managed object context, and any managed object within the store that will
 * be synced must be marked as syncable using the userInfo keys defined below.
 */
@interface SNKPersistentSyncCoordinator : NSObject

- (id)initWithService:(SNKService *)service schema:(NSString *)schema managedObjectContext:(NSManagedObjectContext *)moc changesURL:(NSURL *)changesURL;

@property (strong, readonly) SNKService *service;
@property (copy, readonly) NSString *schema;
@property (strong, readonly) NSManagedObjectContext *managedObjectContext;
@property (copy, readonly) NSURL *changesURL;

@property (unsafe_unretained) id <SNKPersistentSyncCoordinatorDelegate> delegate;

@property (readonly, getter=isSyncing) BOOL syncing;
- (void)sync;
- (void)syncWithOptions:(NSDictionary *)options;
- (void)cancelSync;

// In the next sync session, the entities provided will be slow-synced
- (void)markEntitiesForSlowSync:(NSSet *)entities;

@end


/*!
 * @protocol SNKPersistentSyncCoordinatorDelegate
 * @abstract Adopted by a persistent sync coordinator delegate
 */
@protocol SNKPersistentSyncCoordinatorDelegate <NSObject>

@optional

- (void)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator willStartSyncSession:(SNKSyncSession *)session;
- (void)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator didFinishSyncSession:(SNKSyncSession *)session;
- (void)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator didFailSyncSession:(SNKSyncSession *)session error:(NSError *)error;

- (NSArray *)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator entitiesInSyncSession:(SNKSyncSession *)session;

// Push
- (SNKRecord *)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator willPushRecord:(SNKRecord *)record forManagedObject:(NSManagedObject *)managedObject inSyncSession:(SNKSyncSession *)session;

// Pull
- (SNKRecord *)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator willApplyRecord:(SNKRecord *)record toManagedObject:(NSManagedObject *)managedObject inSyncSession:(SNKSyncSession *)session;
- (void)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator didApplyRecord:(SNKRecord *)record toManagedObject:(NSManagedObject *)managedObject inSyncSession:(SNKSyncSession *)session;

// Conflicts
- (void)persistentSyncCoordinator:(SNKPersistentSyncCoordinator *)coordinator didEncounterConflicts:(NSArray *)conflicts inSyncSession:(SNKSyncSession *)session resolutionHandler:(void (^)(void))resolutionHandler;

@end


/*!
 * @category NSEntityDescription(SNKPersistentSyncCoordinator)
 * @abstract Additions to entities for persistent sync coordination
 */
@interface NSEntityDescription (SNKPersistentSyncCoordinator)

- (NSString *)snk_syncableNameForPropertyName:(NSString *)propertyName;
- (NSString *)snk_propertyNameForSyncableName:(NSString *)syncableName;

@end


/*!
 * @category NSManagedObject(SNKPersistentSyncCoordinator)
 * @abstract Additions to managed objects for persistent sync coordination
 */
@interface NSManagedObject (SNKPersistentSyncCoordinator)

- (SNKProperty *)snk_syncablePropertyForName:(NSString *)name error:(NSError **)outError;
- (BOOL)snk_applySyncableProperty:(SNKProperty *)property error:(NSError **)outError;

@end


/*!
 * @group Entity User Info Keys
 */

/*!
 * @constant SNKEntityIsSyncableKey
 * @abstract Informs a persistent sync coordinator that an entity is syncable.
 *
 * @discussion
 * This key should be set to a truth value (NSNumber or NSString) in an NSEntityDescription's userInfo dictionary.
 */
SYNCKIT_EXTERN NSString * const SNKEntityIsSyncableKey; // @"com.sunflowersw.synckit.syncable";

/*!
 * @constant SNKEntityIdentifierAttributeNameKey
 * @abstract Defines the name of the unique, syncable identifier attribute for an entity class
 *
 * @discussion
 * This key should be set to the name of a modelled attribute (NSString) in every root NSEntityDescription's userInfo dictionary.
 *
 * If no identifier attribute name is specified, a default value of "identifier" is used.
 * If the entity does not define a modelled string attribute with this name an exception is raised when a sync session begins.
 * 
 * A subentity automatically becomes syncable if its superentity defines this key. To disable, set this key to a false truth value in the subentity.
 */
SYNCKIT_EXTERN NSString * const SNKEntityIdentifierAttributeNameKey; // @"com.sunflowersw.synckit.identifier-attribute";


/*!
 * @group Property User Info Keys
 */

/*!
 * @constant SNKPropertyIsSyncableKey
 * @abstract Informs a persistent sync coordinator that a property is syncable.
 *
 * @discussion
 * This key should be set to a truth value (NSNumber or NSString) in an NSPropertyDescription's userInfo dictionary.
 * By default, all properties are ignored and must be explicitly marked for syncing.
 */
SYNCKIT_EXTERN NSString * const SNKPropertyIsSyncableKey; // @"com.sunflowersw.synckit.syncable";

/*!
 * @constant SNKPropertySyncableNameKey
 * @abstract The name of the property in the sync schema to which a managed object property maps
 *
 * @discussion
 * This key should be set to a string value in an NSPropertyDescription's userInfo dictionary.
 * By default, a property's schema counterpart is assumed to have the same name as the property.
 */
SYNCKIT_EXTERN NSString * const SNKPropertySyncableNameKey; // @"com.sunflowersw.synckit.syncable-name";

