//
//  SNKPersistentSyncCoordinator.m
//  SyncKit
//
//  Created by Logan Collins on 3/26/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "SNKPersistentSyncCoordinator.h"
#import "SNKPersistentSyncCoordinator_Private.h"

#import "SNKService.h"
#import "SNKSyncSession.h"

#import "SNKRecord.h"
#import "SNKProperty.h"
#import "SNKAttribute.h"
#import "SNKRelationship.h"

#import "SNKChangeCoordinator.h"
#import "SNKChangeContext.h"
#import "SNKChange.h"
#import "SNKChangeEntity.h"

#import "NSManagedObject+SyncKitAdditions.h"
#import "NSEntityDescription+SyncKitAdditions.h"
#import "NSPropertyDescription+SyncKitAdditions.h"

#import "SyncKitErrors.h"
#import "SNKDebug.h"


NSString * const SNKEntityIsSyncableKey = @"com.sunflowersw.synckit.syncable";
NSString * const SNKEntityIdentifierAttributeNameKey = @"com.sunflowersw.synckit.identifier-attribute";

NSString * const SNKPropertyIsSyncableKey = @"com.sunflowersw.synckit.syncable";
NSString * const SNKPropertyCounterpartNameKey = @"com.sunflowersw.synckit.counterpart";


@implementation SNKPersistentSyncCoordinator {
    SNKChangeCoordinator *_changeCoordinator;
    
    SNKSyncSession *_syncSession;
    SNKChangeContext *_syncSessionChangeContext;
    NSManagedObjectContext *_syncContext;
    NSDictionary *_entitiesByName;
    
    id <NSObject> _activity;
}

- (id)initWithService:(SNKService *)service schema:(NSString *)schema managedObjectContext:(NSManagedObjectContext *)moc changesURL:(NSURL *)changesURL {
    NSParameterAssert(service != nil);
    NSParameterAssert(moc != nil);
    NSParameterAssert(changesURL != nil);
    
    self = [super init];
    if (self) {
        self.service = service;
        self.schema = schema;
        self.managedObjectContext = moc;
        self.changesURL = changesURL;
        
        _changeCoordinator = [[SNKChangeCoordinator alloc] initWithURL:changesURL];
        
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(managedObjectContextWillSave:) name:NSManagedObjectContextWillSaveNotification object:moc];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(managedObjectContextDidSave:) name:NSManagedObjectContextDidSaveNotification object:moc];
    }
    return self;
}

- (void)dealloc {
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}


#pragma mark -
#pragma mark Notifications

- (void)managedObjectContextWillSave:(NSNotification *)notification {
    NSManagedObjectContext *context = [notification object];
    
    NSSet *insertedObjects = [context insertedObjects];
    NSSet *updatedObjects = [context updatedObjects];
    NSSet *deletedObjects = [context deletedObjects];
    
    if ([insertedObjects count] > 0 || [updatedObjects count] > 0 || [deletedObjects count] > 0) {
        SNKDebugLog(@"SyncKit observing managed object changes:\ninserted=%@,\nupdated=%@,\ndeleted=%@", insertedObjects, updatedObjects, deletedObjects);
        
        SNKChangeContext *changeContext = [[SNKChangeContext alloc] initWithChangeCoordinator:_changeCoordinator];
        [changeContext performBlockAndWait:^{
            // Build a transaction log for syncing
            for (id object in insertedObjects) {
                if ([[object entity] snk_isSyncable]) {
                    NSString *identifier = [object snk_syncableIdentifier];
                    NSString *entityName = [[object entity] name];
                    if (identifier != nil) {
                        [changeContext insertedRecordWithIdentifier:identifier entity:entityName attributes:nil];
                    }
                }
            }
            
            for (id object in updatedObjects) {
                if ([[object entity] snk_isSyncable]) {
                    NSString *identifier = [object snk_syncableIdentifier];
                    NSString *entityName = [[object entity] name];
                    
                    NSDictionary *syncableAttributesByName = [[object entity] snk_syncablePropertiesByName];
                    NSMutableSet *syncableAttributes = [NSMutableSet setWithCapacity:[syncableAttributesByName count]];
                    
                    for (NSString *syncableName in syncableAttributesByName) {
                        NSPropertyDescription *description = [syncableAttributesByName objectForKey:syncableName];
                        NSString *name = [description name];
                        [syncableAttributes addObject:name];
                    }
                    
                    NSSet *changeKeys = [NSSet setWithArray:[[object changedValues] allKeys]];
                    [syncableAttributes intersectSet:changeKeys];
                    
                    if (identifier != nil && [syncableAttributes count] > 0) {
                        [changeContext updatedRecordWithIdentifier:identifier entity:entityName attributes:syncableAttributes];
                    }
                }
            }
            
            for (id object in deletedObjects) {
                if ([[object entity] snk_isSyncable]) {
                    NSString *identifier = [object snk_syncableIdentifier];
                    NSString *entityName = [[object entity] name];
                    if (identifier != nil) {
                        [changeContext deletedRecordWithIdentifier:identifier entity:entityName];
                    }
                }
            }
            
            NSError *error = nil;
            if (![changeContext commit:&error]) {
                NSLog(@"Error while committing change context: %@", error);
            }
        }];
    }
}

- (void)managedObjectContextDidSave:(NSNotification *)notification {
    
}


#pragma mark -
#pragma mark Syncing

- (void)sync {
    [self syncWithOptions:nil];
}

- (void)syncWithOptions:(NSDictionary *)options {
    if (self.syncing) {
        return;
    }
    
    self.syncing = YES;
    
    if ([options[SNKSyncSessionSlowSyncOption] boolValue]) {
        [self markEntitiesForSlowSync:nil];
    }
    
    _syncSession = [self.service beginSyncSessionForSchema:self.schema delegate:self options:options];
}

- (void)cancelSync {
    [_syncSession cancel];
}

- (void)syncContextDidSave:(NSNotification *)notification {
    [self.managedObjectContext performBlock:^{
        [self.managedObjectContext mergeChangesFromContextDidSaveNotification:notification];
    }];
}

- (void)markEntitiesForSlowSync:(NSSet *)entities {
    if (entities == nil) {
        // Mark all entities for slow sync
        entities = [NSSet setWithArray:self.managedObjectContext.persistentStoreCoordinator.managedObjectModel.entities];
    }
    
    NSMutableDictionary *entitiesToSync = [NSMutableDictionary dictionaryWithCapacity:[entities count]];
    for (NSEntityDescription *entity in entities) {
        NSString *name = [entity name];
        if ([entity snk_isSyncable] && ![entity isAbstract]) {
            entitiesToSync[name] = entity;
        }
    }
    
    SNKChangeContext *context = [[SNKChangeContext alloc] initWithChangeCoordinator:_changeCoordinator];
    [context performBlockAndWait:^{
        for (NSString *entityName in entitiesToSync) {
            SNKChangeEntity *entity = [context entityWithName:entityName create:YES];
            entity.shouldPerformSlowSync = YES;
        }
        
        NSError *error = nil;
        if (![context commit:&error]) {
            NSLog(@"Error while committing change context: %@", error);
        }
    }];
}


#pragma mark -
#pragma mark Sync Session delegate

- (void)syncSessionWillStart:(SNKSyncSession *)session {
    if ([self.delegate respondsToSelector:@selector(persistentSyncCoordinator:willStartSyncSession:)]) {
        [self.delegate persistentSyncCoordinator:self willStartSyncSession:session];
    }
    
#if TARGET_OS_IPHONE
#if __IPHONE_OS_VERSION_MAX_ALLOWED >= 70000
    if ([NSProcessInfo instancesRespondToSelector:@selector(beginActivityWithOptions:reason:)]) {
        _activity = [[NSProcessInfo processInfo] beginActivityWithOptions:(NSActivityBackground) reason:@"Syncing persistent stores"];
    }
#endif
#else
    if ([NSProcessInfo instancesRespondToSelector:@selector(beginActivityWithOptions:reason:)]) {
        _activity = [[NSProcessInfo processInfo] beginActivityWithOptions:(NSActivityBackground) reason:@"Syncing persistent stores"];
    }
    else {
        if ([NSProcessInfo instancesRespondToSelector:@selector(disableSuddenTermination)]) {
            [[NSProcessInfo processInfo] disableSuddenTermination];
        }
        if ([NSProcessInfo instancesRespondToSelector:@selector(disableAutomaticTermination:)]) {
            [[NSProcessInfo processInfo] disableAutomaticTermination:@"Syncing persistent stores"];
        }
    }
#endif
    
    // Create a private, concurrent sync context
    _syncContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
    _syncContext.persistentStoreCoordinator = self.managedObjectContext.persistentStoreCoordinator;
    _syncContext.mergePolicy = [[NSMergePolicy alloc] initWithMergeType:NSMergeByPropertyStoreTrumpMergePolicyType];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(syncContextDidSave:) name:NSManagedObjectContextDidSaveNotification object:_syncContext];
    
    _syncSessionChangeContext = [[SNKChangeContext alloc] initWithChangeCoordinator:_changeCoordinator];
    
    NSArray *entities = nil;
    if ([self.delegate respondsToSelector:@selector(persistentSyncCoordinator:entitiesInSyncSession:)]) {
        entities = [self.delegate persistentSyncCoordinator:self entitiesInSyncSession:session];
    }
    if (entities == nil) {
        entities = self.managedObjectContext.persistentStoreCoordinator.managedObjectModel.entities;
    }
    
    NSMutableDictionary *entitiesToSync = [NSMutableDictionary dictionaryWithCapacity:[entities count]];
    for (NSEntityDescription *entity in entities) {
        NSString *name = [entity name];
        
        if ([entity snk_isSyncable] && ![entity isAbstract]) {
            entitiesToSync[name] = entity;
        }
    }
    _entitiesByName = entitiesToSync;
}

- (BOOL)syncSessionShouldFinalize:(SNKSyncSession *)session error:(NSError *__autoreleasing *)outError {
    __block NSError *localError = nil;
    
    [_syncSessionChangeContext performBlockAndWait:^{
        NSDate *timestamp = [NSDate date];
        for (NSString *entityName in _entitiesByName) {
            SNKChangeEntity *entity = [_syncSessionChangeContext entityWithName:entityName create:YES];
            entity.lastSyncDate = timestamp;
            entity.shouldPerformSlowSync = NO;
        }
        
        [_syncSessionChangeContext commit:&localError];
    }];
    
    if (localError == nil) {
        [_syncContext performBlockAndWait:^{
            [_syncContext save:&localError];
        }];
    }
    
    if (outError != NULL) {
        *outError = localError;
    }
    
    return (localError == nil);
}

- (void)syncSessionDidFinish:(SNKSyncSession *)session {
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSManagedObjectContextDidSaveNotification object:_syncContext];
    
    _syncSession = nil;
    _syncContext = nil;
    _entitiesByName = nil;
    _syncSessionChangeContext = nil;
    
    self.syncing = NO;
    
#if TARGET_OS_IPHONE
#if __IPHONE_OS_VERSION_MAX_ALLOWED >= 70000
    if ([NSProcessInfo instancesRespondToSelector:@selector(endActivity:)]) {
        if (_activity != nil) {
            [[NSProcessInfo processInfo] endActivity:_activity];
        }
    }
#endif
#else
    if ([NSProcessInfo instancesRespondToSelector:@selector(endActivity:)]) {
        if (_activity != nil) {
            [[NSProcessInfo processInfo] endActivity:_activity];
        }
    }
    else {
        if ([NSProcessInfo instancesRespondToSelector:@selector(enableSuddenTermination)]) {
            [[NSProcessInfo processInfo] enableSuddenTermination];
        }
        if ([NSProcessInfo instancesRespondToSelector:@selector(enableAutomaticTermination:)]) {
            [[NSProcessInfo processInfo] enableAutomaticTermination:@"Syncing persistent stores"];
        }
    }
#endif
    
    if ([self.delegate respondsToSelector:@selector(persistentSyncCoordinator:didFinishSyncSession:)]) {
        [self.delegate persistentSyncCoordinator:self didFinishSyncSession:session];
    }
}

- (void)syncSession:(SNKSyncSession *)session didFailWithError:(NSError *)error {
    [_syncSessionChangeContext performBlockAndWait:^{
        [_syncSessionChangeContext rollback];
    }];
    
    [_syncContext performBlockAndWait:^{
        [_syncContext rollback];
        for (NSManagedObject *object in [_syncContext registeredObjects]) {
            // Turn all objects into faults
            [_syncContext refreshObject:object mergeChanges:NO];
        }
    }];
    
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSManagedObjectContextDidSaveNotification object:_syncContext];
    
    _syncSession = nil;
    _syncContext = nil;
    _entitiesByName = nil;
    _syncSessionChangeContext = nil;
    
    self.syncing = NO;
    
#if TARGET_OS_IPHONE
#if __IPHONE_OS_VERSION_MAX_ALLOWED >= 70000
    if ([NSProcessInfo instancesRespondToSelector:@selector(endActivity:)]) {
        if (_activity != nil) {
            [[NSProcessInfo processInfo] endActivity:_activity];
        }
    }
#endif
#else
    if ([NSProcessInfo instancesRespondToSelector:@selector(endActivity:)]) {
        if (_activity != nil) {
            [[NSProcessInfo processInfo] endActivity:_activity];
        }
    }
    else {
        if ([NSProcessInfo instancesRespondToSelector:@selector(enableSuddenTermination)]) {
            [[NSProcessInfo processInfo] enableSuddenTermination];
        }
        if ([NSProcessInfo instancesRespondToSelector:@selector(enableAutomaticTermination:)]) {
            [[NSProcessInfo processInfo] enableAutomaticTermination:@"Syncing persistent stores"];
        }
    }
#endif
    
    if ([self.delegate respondsToSelector:@selector(persistentSyncCoordinator:didFailSyncSession:error:)]) {
        [self.delegate persistentSyncCoordinator:self didFailSyncSession:session error:error];
    }
}

- (NSArray *)syncSessionEntitiesToSync:(SNKSyncSession *)session {
    return [_entitiesByName allKeys];
}

- (NSDictionary *)syncSession:(SNKSyncSession *)session recordsForEntity:(NSString *)entityName moreComing:(BOOL *)moreComing error:(NSError *__autoreleasing *)outError {
    __block NSError *error = nil;
    __block NSMutableDictionary *records = [NSMutableDictionary dictionary];
    
    BOOL delegateRespondsToWillPushRecord = [self.delegate respondsToSelector:@selector(persistentSyncCoordinator:willPushRecord:forManagedObject:inSyncSession:)];
    
    [_syncContext performBlockAndWait:^{
        NSMutableArray *errors = [NSMutableArray array];
        
        NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:entityName];
        NSArray *results = [_syncContext executeFetchRequest:request error:nil];
        
        for (NSManagedObject *object in results) {
            if (![[[object entity] name] isEqualToString:entityName]) {
                continue;
            }
            
            NSString *identifier = [object snk_syncableIdentifier];
            
            SNKRecord *record = [[SNKRecord alloc] init];
            record.identifier = identifier;
            record.entityName = [[object entity] name];
            
            // Properties
            NSDictionary *propertyNames = [[object entity] snk_syncablePropertiesByName];
            NSMutableArray *properties = [NSMutableArray arrayWithCapacity:[propertyNames count]];
            for (NSString *propertyName in propertyNames) {
                NSError *propertyError = nil;
                SNKProperty *property = [object snk_syncablePropertyForName:propertyName error:&propertyError];
                
                if (property != nil) {
                    [properties addObject:property];
                }
                else if (propertyError != nil) {
                    [errors addObject:propertyError];
                }
            }
            record.properties = properties;
            
            if (delegateRespondsToWillPushRecord) {
                record = [self.delegate persistentSyncCoordinator:self willPushRecord:record forManagedObject:object inSyncSession:session];
            }
            
            if (record != nil) {
                records[identifier] = record;
            }
        }
        
        // Errors
        if ([errors count] > 0) {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Multiple validation errors occurred", nil), NSLocalizedDescriptionKey,
                                      NSLocalizedString(@"Errors were encountered while applying sync values.", nil), NSLocalizedFailureReasonErrorKey,
                                      errors, @"Errors",
                                      nil];
            error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKValidationError userInfo:userInfo];
        }
    }];
    
    // Discard change objects during slow-sync
    [_syncSessionChangeContext performBlockAndWait:^{
        [_syncSessionChangeContext deleteAllChangesForEntity:[_syncSessionChangeContext entityWithName:entityName create:NO]];
    }];
    
    if ([records count] > 0) {
        SNKDebugLog(@"Pushing records for entity: %@, %@", entityName, records);
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return records;
}

- (BOOL)syncSession:(SNKSyncSession *)session shouldPerformFastSyncForEntity:(NSString *)entityName {
    __block BOOL slow = NO;
    [_syncSessionChangeContext performBlockAndWait:^{
        SNKChangeEntity *entity = [_syncSessionChangeContext entityWithName:entityName create:NO];
        slow = (entity != nil && !entity.shouldPerformSlowSync);
    }];
    return slow;
}

- (NSDictionary *)syncSession:(SNKSyncSession *)session recordsChangedForEntity:(NSString *)entityName moreComing:(BOOL *)moreComing error:(NSError *__autoreleasing *)outError {
    __block NSError *error = nil;
    __block NSMutableDictionary *records = [NSMutableDictionary dictionary];
    
    __block NSMutableArray *changeDicts = nil;
    __block NSMutableArray *identifiers = nil;
    
    BOOL delegateRespondsToWillPushRecord = [self.delegate respondsToSelector:@selector(persistentSyncCoordinator:willPushRecord:forManagedObject:inSyncSession:)];
    
    NSEntityDescription *entity = _entitiesByName[entityName];
    NSString *identityAttributeName = [entity snk_syncableIdentifierAttributeName];
    
    [_syncSessionChangeContext performBlockAndWait:^{
        // Discard change objects during slow-sync
        NSPredicate *predicate = [NSPredicate predicateWithFormat:@"changeEntity.name == %@", entityName];
        NSArray *changes = [_syncSessionChangeContext changesMatchingPredicate:predicate];
        
        changeDicts = [NSMutableArray arrayWithCapacity:[changes count]];
        identifiers = [NSMutableArray arrayWithCapacity:[changes count]];
        
        for (SNKChange *change in changes) {
            NSMutableDictionary *changeDict = [NSMutableDictionary dictionary];
            changeDict[@"recordIdentifier"] = change.identifier;
            changeDict[@"entityName"] = change.changeEntity.name;
            changeDict[@"type"] = @(change.type);
            if (change.date != nil) {
                changeDict[@"date"] = change.date;
            }
            if (change.attributes != nil) {
                changeDict[@"changedAttributes"] = change.attributes;
            }
            [changeDicts addObject:changeDict];
            
            [identifiers addObject:change.identifier];
            
            [_syncSessionChangeContext deleteChange:change];
        }
    }];
    
    if ([changeDicts count] > 0) {
        [_syncContext performBlockAndWait:^{
            NSMutableArray *errors = [NSMutableArray array];
            
            NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:entityName];
            request.predicate = [NSPredicate predicateWithFormat:@"%K IN %@", identityAttributeName, identifiers];
            NSArray *objects = [_syncContext executeFetchRequest:request error:nil];
            NSMutableDictionary *objectsByIdentifier = [NSMutableDictionary dictionaryWithCapacity:[objects count]];
            for (NSManagedObject *object in objects) {
                objectsByIdentifier[[object snk_syncableIdentifier]] = object;
            }
            
            for (NSDictionary *changeDict in changeDicts) {
                NSString *identifier = changeDict[@"recordIdentifier"];
                SNKChangeType type = [changeDict[@"type"] integerValue];
                NSDate *date = changeDict[@"date"];
//                NSSet *changedAttributes = changeDict[@"changedAttributes"];  NSSet *changedAttributes = change.attributes;
                
                SNKRecord *record = records[identifier];
                if (record == nil) {
                    record = [[SNKRecord alloc] init];
                    record.identifier = identifier;
                    record.entityName = entityName;
                }
                
                NSManagedObject *object = nil;
                
                if (type == SNKChangeTypeInsert || type == SNKChangeTypeUpdate) {
                    object = objectsByIdentifier[identifier];
                    if (object != nil) {
                        record.modificationDate = date;
                        
                        // Properties
                        NSDictionary *propertyNames = nil;
                        if (type == SNKChangeTypeInsert) {
                            // Sync all properties for inserted objects
                            propertyNames = [entity snk_syncablePropertiesByName];
                        }
                        else if (type == SNKChangeTypeUpdate) {
//                            if (changedAttributes != nil) {
//                                // Sync only changed objects
//                                propertyNames = changedAttributes;
//                            }
//                            else {
                                // Sync all attributes
                                propertyNames = [entity snk_syncablePropertiesByName];
//                            }
                        }
                        
                        NSMutableArray *properties = [NSMutableArray arrayWithCapacity:[propertyNames count]];
                        for (NSString *propertyName in propertyNames) {
                            NSError *propertyError = nil;
                            SNKProperty *property = [object snk_syncablePropertyForName:propertyName error:&propertyError];
                            
                            if (property != nil) {
                                [properties addObject:property];
                            }
                            else if (propertyError != nil) {
                                [errors addObject:propertyError];
                            }
                        }
                        [record addProperties:properties];
                        
                        records[identifier] = record;
                    }
                }
                else if (type == SNKChangeTypeDelete) {
                    record.deleted = YES;
                    record.deletionDate = date;
                }
                
                if (delegateRespondsToWillPushRecord) {
                    record = [self.delegate persistentSyncCoordinator:self willPushRecord:record forManagedObject:object inSyncSession:session];
                }
                
                if (record != nil) {
                    records[identifier] = record;
                }
            }
            
            // Errors
            if ([errors count] > 0) {
                NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                          NSLocalizedString(@"Multiple validation errors occurred", nil), NSLocalizedDescriptionKey,
                                          NSLocalizedString(@"Errors were encountered while applying sync values.", nil), NSLocalizedFailureReasonErrorKey,
                                          errors, @"Errors",
                                          nil];
                error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKValidationError userInfo:userInfo];
            }
        }];
        
        if ([records count] > 0) {
            SNKDebugLog(@"Pushing changed records for entity: %@, %@", entityName, records);
        }
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return records;
}

- (BOOL)syncSession:(SNKSyncSession *)session applyRecords:(NSDictionary *)records moreComing:(BOOL)moreComing error:(NSError *__autoreleasing *)outError {
    __block NSError *error = nil;
    
    SNKDebugLog(@"Applying records: %@", records);
    
    BOOL delegateRespondsToWillApplyRecord = [self.delegate respondsToSelector:@selector(persistentSyncCoordinator:willApplyRecord:toManagedObject:inSyncSession:)];
    BOOL delegateRespondsToDidApplyRecord = [self.delegate respondsToSelector:@selector(persistentSyncCoordinator:didApplyRecord:toManagedObject:inSyncSession:)];
    
    [_syncContext performBlockAndWait:^{
        NSMutableArray *errors = [NSMutableArray array];
        
        // Prefetch all existing objects
        NSMutableDictionary *objectsByIdentifier = [NSMutableDictionary dictionaryWithCapacity:[records count]];
        
        NSMutableDictionary *identifiersByEntityName = [NSMutableDictionary dictionaryWithCapacity:[_entitiesByName count]];
        for (NSString *identifier in records) {
            SNKRecord *record = records[identifier];
            NSString *entityName = record.entityName;
            
            NSMutableArray *identifiers = identifiersByEntityName[entityName];
            if (identifiers == nil) {
                identifiers = [NSMutableArray arrayWithCapacity:[records count]];
                identifiersByEntityName[entityName] = identifiers;
            }
            
            [identifiers addObject:identifier];
        }
        
        for (NSString *entityName in identifiersByEntityName) {
            NSArray *identifiers = identifiersByEntityName[entityName];
            
            NSEntityDescription *entity = _entitiesByName[entityName];
            NSString *identityAttributeName = [entity snk_syncableIdentifierAttributeName];
            
            NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:entityName];
            request.predicate = [NSPredicate predicateWithFormat:@"%K IN %@", identityAttributeName, identifiers];
            NSArray *results = [_syncContext executeFetchRequest:request error:nil];
            
            for (NSManagedObject *object in results) {
                objectsByIdentifier[[object snk_syncableIdentifier]] = object;
            }
        }
        
        
        // First pass: Inserted records
        // Ensure all pulled records exist, creating if necessary
        for (NSString *identifier in records) {
            SNKRecord *record = records[identifier];
            
            if (record.deleted) {
                // Don't insert deleted records
                continue;
            }
            
            NSManagedObject *object = objectsByIdentifier[identifier];
            if (object == nil) {
                object = [NSEntityDescription insertNewObjectForEntityForName:record.entityName inManagedObjectContext:_syncContext];
                [object snk_setSyncableIdentifier:identifier];
                objectsByIdentifier[identifier] = object;
            }
        }
        
        
        // Second pass: Updated records
        // Update all pulled record properties
        for (NSString *identifier in records) {
            SNKRecord *record = records[identifier];
            
            if (record.deleted) {
                // Don't update deleted records
                continue;
            }
            
            NSManagedObject *object = objectsByIdentifier[identifier];
            if (object == nil) {
                continue;
            }
            
            if (delegateRespondsToWillApplyRecord) {
                record = [self.delegate persistentSyncCoordinator:self willApplyRecord:record toManagedObject:object inSyncSession:session];
                if (record == nil) {
                    continue;
                }
            }
            
            NSDictionary *propertyNames = [[object entity] snk_syncablePropertiesByName];
            NSDictionary *properties = [record propertiesByName];
            
            // Update properties
            for (NSString *propertyName in propertyNames) {
                SNKProperty *property = properties[propertyName];
                if (property != nil) {
                    NSError *propertyError = nil;
                    if (![object snk_applySyncableProperty:property error:&propertyError]) {
                        if (propertyError != nil) {
                            [errors addObject:propertyError];
                        }
                    }
                }
            }
            
            if (delegateRespondsToDidApplyRecord) {
                [self.delegate persistentSyncCoordinator:self didApplyRecord:record toManagedObject:object inSyncSession:session];
            }
        }
        
        
        // Third pass: Deleted records
        // Remove records marked as deleted
        for (NSString *identifier in records) {
            SNKRecord *record = records[identifier];
            
            if (record.deleted) {
                NSManagedObject *object = objectsByIdentifier[identifier];
                if (object != nil) {
                    if (delegateRespondsToWillApplyRecord) {
                        record = [self.delegate persistentSyncCoordinator:self willApplyRecord:record toManagedObject:object inSyncSession:session];
                        if (record == nil) {
                            continue;
                        }
                    }
                    
                    [_syncContext deleteObject:object];
                    
                    if (delegateRespondsToDidApplyRecord) {
                        [self.delegate persistentSyncCoordinator:self didApplyRecord:record toManagedObject:object inSyncSession:session];
                    }
                }
            }
        }
        
        
        // Errors
        if ([errors count] > 0) {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Multiple validation errors occurred", nil), NSLocalizedDescriptionKey,
                                      NSLocalizedString(@"Errors were encountered while applying sync values.", nil), NSLocalizedFailureReasonErrorKey,
                                      errors, @"Errors",
                                      nil];
            error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKValidationError userInfo:userInfo];
        }
    }];
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return (error == nil);
}

@end


@implementation NSEntityDescription (SNKPersistentSyncCoordinator)

- (NSString *)snk_syncableNameForPropertyName:(NSString *)propertyName {
    NSDictionary *properties = [self propertiesByName];
    NSPropertyDescription *property = [properties objectForKey:propertyName];
    return [property snk_syncableName];
}

- (NSString *)snk_propertyNameForSyncableName:(NSString *)syncableName {
    NSDictionary *syncableProperties = [self snk_syncablePropertiesByName];
    NSPropertyDescription *property = [syncableProperties objectForKey:syncableName];
    return [property name];
}

@end


@implementation NSManagedObject (SNKPersistentSyncCoordinator)

- (NSDictionary *)snk_objectsWithIdentifiers:(NSArray *)identifiers entityName:(NSString *)entityName {
    __block NSMutableDictionary *objects = nil;
    
    NSManagedObjectContext *context = [self managedObjectContext];
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:entityName];
    
    NSString *syncableIdentifierAttributeName = [[self entity] snk_syncableIdentifierAttributeName];
    request.predicate = [NSPredicate predicateWithFormat:@"%K IN %@", syncableIdentifierAttributeName, identifiers];
    
    NSArray *results = [context executeFetchRequest:request error:nil];
    objects = [NSMutableDictionary dictionaryWithCapacity:[results count]];
    for (NSManagedObject *object in results) {
        objects[[object snk_syncableIdentifier]] = object;
    }
    
    return objects;
}

- (SNKProperty *)snk_syncablePropertyForName:(NSString *)name error:(NSError **)outError {
    SNKProperty *property = nil;
    NSError *error = nil;
    
    NSEntityDescription *entity = [self entity];
    id objectProperty = [[entity snk_syncablePropertiesByName] objectForKey:name];
    if (objectProperty == nil) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), name, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
    }
    
    NSString *objectPropertyName = [objectProperty name];
    
    if ([objectProperty isKindOfClass:[NSAttributeDescription class]]) {
        NSAttributeType type = [objectProperty attributeType];
        id value = [self valueForKey:objectPropertyName];
        
        SNKAttribute *attribute = [[SNKAttribute alloc] init];
        attribute.name = name;
        
        if (type == NSInteger16AttributeType
            || type == NSInteger32AttributeType
            || type == NSInteger64AttributeType) {
            attribute.type = SNKAttributeTypeInteger;
        }
        else if (type == NSFloatAttributeType
                 || type == NSDoubleAttributeType) {
            attribute.type = SNKAttributeTypeFloat;
        }
        else if (type == NSDecimalAttributeType) {
            attribute.type = SNKAttributeTypeDecimal;
        }
        else if (type == NSBooleanAttributeType) {
            attribute.type = SNKAttributeTypeBoolean;
        }
        else if (type == NSStringAttributeType) {
            attribute.type = SNKAttributeTypeString;
        }
        else if (type == NSDateAttributeType) {
            attribute.type = SNKAttributeTypeDate;
        }
        else if (type == NSBinaryDataAttributeType) {
            attribute.type = SNKAttributeTypeBinaryData;
        }
        else if (type == NSTransformableAttributeType) {
            attribute.type = SNKAttributeTypeBinaryData;
            
            NSString *valueTransformerName = [objectProperty valueTransformerName];
            if (value == nil) {
                
            }
            else if (valueTransformerName == nil) {
                value = [NSKeyedArchiver archivedDataWithRootObject:value];
            }
            else {
                NSValueTransformer *transformer = [NSValueTransformer valueTransformerForName:valueTransformerName];
                value = [transformer transformedValue:value];
            }
        }
        else {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), name, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
        }
        
        if (error == nil) {
            attribute.value = value;
            
            property = attribute;
        }
    }
    else if ([objectProperty isKindOfClass:[NSRelationshipDescription class]]) {
        NSEntityDescription *destinationEntity = [objectProperty destinationEntity];
        if ([destinationEntity snk_isSyncable]) {
            SNKRelationship *relationship = [[SNKRelationship alloc] init];
            relationship.name = name;
            relationship.toMany = [objectProperty isToMany];
            relationship.ordered = [objectProperty isOrdered];
            
            if ([objectProperty isToMany]) {
                id value = [self valueForKey:objectPropertyName];
                if ([objectProperty isOrdered]) {
                    NSOrderedSet *relatedObjects = value;
                    NSMutableArray *array = [NSMutableArray arrayWithCapacity:[relatedObjects count]];
                    for (NSManagedObject *relatedObject in relatedObjects) {
                        [array addObject:[relatedObject snk_syncableIdentifier]];
                    }
                    relationship.identifiers = array;
                }
                else {
                    NSSet *relatedObjects = value;
                    NSMutableArray *array = [NSMutableArray arrayWithCapacity:[relatedObjects count]];
                    for (NSManagedObject *relatedObject in relatedObjects) {
                        [array addObject:[relatedObject snk_syncableIdentifier]];
                    }
                    relationship.identifiers = array;
                }
            }
            else {
                NSManagedObject *relatedObject = [self valueForKey:objectPropertyName];
                if (relatedObject != nil) {
                    relationship.identifiers = @[[relatedObject snk_syncableIdentifier]];
                }
            }
            
            property = relationship;
        }
        else {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to a relationship of entity '%@'", nil), name, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
        }
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return property;
}

- (BOOL)snk_applySyncableProperty:(SNKProperty *)property error:(NSError **)outError {
    BOOL success = NO;
    NSError *error = nil;
    
    NSString *name = property.name;
    
    NSEntityDescription *entity = [self entity];
    id objectProperty = [[entity snk_syncablePropertiesByName] objectForKey:name];
    if (objectProperty == nil) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), property, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
        success = NO;
    }
    
    NSString *objectPropertyName = [objectProperty name];
    
    if ([objectProperty isKindOfClass:[NSAttributeDescription class]] && [property isKindOfClass:[SNKAttribute class]]) {
        // Attribute
        NSAttributeType type = [objectProperty attributeType];
        //        SNKAttributeType valueType = [(SNKAttribute *)property type];
        
        id propertyValue = nil;
        id newValue = [(SNKAttribute *)property value];
        
        BOOL valid = NO;
        
        if (type == NSInteger16AttributeType
            || type == NSInteger32AttributeType
            || type == NSInteger64AttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSNumber class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSFloatAttributeType
                 || type == NSDoubleAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSNumber class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSDecimalAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSDecimalNumber class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSBooleanAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSNumber class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSStringAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSString class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSDateAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSDate class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSBinaryDataAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSData class]]) {
                propertyValue = newValue;
                valid = YES;
            }
        }
        else if (type == NSTransformableAttributeType) {
            if (newValue == nil || [newValue isKindOfClass:[NSData class]]) {
                NSString *valueTransformerName = [(NSAttributeDescription *)objectProperty valueTransformerName];
                if (newValue == nil) {
                    propertyValue = newValue;
                }
                else if (valueTransformerName == nil) {
                    propertyValue = [NSKeyedUnarchiver unarchiveObjectWithData:newValue];
                }
                else {
                    NSValueTransformer *transformer = [NSValueTransformer valueTransformerForName:valueTransformerName];
                    propertyValue = [transformer reverseTransformedValue:newValue];
                }
                valid = YES;
            }
        }
        
        if (valid) {
            if (propertyValue == nil && ![objectProperty isOptional] && [(NSAttributeDescription *)objectProperty defaultValue] != nil) {
                // Reset to default value
                propertyValue = [(NSAttributeDescription *)objectProperty defaultValue];
            }
            
            if (propertyValue == nil && ![objectProperty isOptional]) {
                NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                          NSLocalizedString(@"Attribute cannot be empty", nil), NSLocalizedDescriptionKey,
                                          [NSString stringWithFormat:NSLocalizedString(@"Required attribute '%@' of entity '%@' cannot be empty.", nil), name, [entity name]], NSLocalizedDescriptionKey,
                                          nil];
                error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKMissingRequiredValueError userInfo:userInfo];
                success = NO;
            }
            else {
                // Validate and set
                if ([self validateValue:&propertyValue forKey:objectPropertyName error:&error]) {
                    [self setValue:propertyValue forKey:objectPropertyName];
                    success = YES;
                }
            }
        }
        else {
            // Unsupported attribute type
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map attribute '%@' to an attribute of entity '%@'", nil), property, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
            success = NO;
        }
    }
    else if ([objectProperty isKindOfClass:[NSRelationshipDescription class]] && [property isKindOfClass:[SNKRelationship class]]) {
        // Relationship
        SNKRelationship *relationship = (SNKRelationship *)property;
        NSEntityDescription *destinationEntity = [objectProperty destinationEntity];
        
        NSArray *relatedIdentifiers = relationship.identifiers;
        NSDictionary *relatedObjects = [self snk_objectsWithIdentifiers:relatedIdentifiers entityName:[destinationEntity name]];
        
        id newValue = nil;
        
        if ([(NSRelationshipDescription *)objectProperty isToMany]) {
            if ([(NSRelationshipDescription *)objectProperty isOrdered]) {
                NSMutableOrderedSet *set = [NSMutableOrderedSet orderedSetWithCapacity:[relatedIdentifiers count]];
                for (NSString *relatedIdentifier in relatedIdentifiers) {
                    NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                    if (relatedObject != nil) {
                        [set addObject:relatedObject];
                    }
                }
                newValue = set;
            }
            else {
                NSMutableSet *set = [NSMutableSet setWithCapacity:[relatedIdentifiers count]];
                for (NSString *relatedIdentifier in relatedIdentifiers) {
                    NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                    if (relatedObject != nil) {
                        [set addObject:relatedObject];
                    }
                }
                newValue = set;
            }
        }
        else {
            if ([relatedIdentifiers count] > 0) {
                NSString *relatedIdentifier = relatedIdentifiers[0];
                NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                newValue = relatedObject;
            }
        }
        
        if ([self validateValue:&newValue forKey:objectPropertyName error:&error]) {
            [self setValue:newValue forKey:objectPropertyName];
            success = YES;
        }
    }
    else {
        // Unknown
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to a property of entity '%@'", nil), property, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:SNKSyncKitErrorDomain code:SNKUnsupportedPropertyMappingError userInfo:userInfo];
        success = NO;
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return success;
}

@end
