//
//  SNKAttribute.m
//  SyncKit
//
//  Created by Logan Collins on 12/19/12.
//  Copyright (c) 2012 Sunflower Softworks. All rights reserved.
//

#import "SNKAttribute.h"
#import "SNKAttribute_Private.h"

#import "SNKProperty_Private.h"

#import "NSData+SyncKitAdditions.h"


@implementation SNKAttribute

+ (SNKAttribute *)attributeWithName:(NSString *)name type:(SNKAttributeType)type value:(id)value {
    SNKAttribute *attribute = [[self alloc] init];
    attribute.name = name;
    attribute.type = type;
    attribute.value = value;
    return attribute;
}

- (id)initWithDictionary:(NSDictionary *)dictionary {
    self = [super initWithDictionary:dictionary];
    if (self) {
        NSString *typeString = dictionary[@"type"];
        self.type = SNKAttributeTypeFromString(typeString);
        
        id value = dictionary[@"value"];
        
        if (value == nil || [value isKindOfClass:[NSNull class]]) {
            self.value = nil;
        }
        else if (self.type == SNKAttributeTypeInteger
            || self.type == SNKAttributeTypeFloat
            || self.type == SNKAttributeTypeBoolean) {
            if ([value isKindOfClass:[NSNumber class]]) {
                self.value = value;
            }
        }
        else if (self.type == SNKAttributeTypeDecimal) {
            if ([value isKindOfClass:[NSString class]]) {
                static NSNumberFormatter *numberFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    numberFormatter = [[NSNumberFormatter alloc] init];
                    [numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
                    [numberFormatter setGeneratesDecimalNumbers:YES];
                });
                self.value = [numberFormatter numberFromString:value];
            }
        }
        else if (self.type == SNKAttributeTypeString) {
            if ([value isKindOfClass:[NSString class]]) {
                self.value = value;
            }
        }
        else if (self.type == SNKAttributeTypeDate) {
            if ([value isKindOfClass:[NSString class]]) {
                static NSDateFormatter *dateFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    dateFormatter = [[NSDateFormatter alloc] init];
                    [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
                    
                    NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
                    [dateFormatter setLocale:locale];
                });
                self.value = [dateFormatter dateFromString:value];
            }
        }
        else if (self.type == SNKAttributeTypeBinaryData) {
            if ([value isKindOfClass:[NSString class]]) {
                NSData *data = [NSData snk_dataWithBase64String:value];
                self.value = data;
            }
        }
    }
    return self;
}

- (id)copyWithZone:(NSZone *)zone {
    SNKAttribute *attribute = [super copyWithZone:zone];
    attribute.type = self.type;
    attribute.value = self.value;
    return attribute;
}

- (NSString *)description {
    return [NSString stringWithFormat:@"%@ { name='%@', type='%@', value='%@' }", [super description], self.name, SNKStringFromAttributeType(self.type), self.value];
}

- (NSDictionary *)dictionaryRepresentation {
    NSMutableDictionary *dictionary = [[super dictionaryRepresentation] mutableCopy];
    
    dictionary[@"type"] = SNKStringFromAttributeType(self.type);
    
    if (self.value != nil) {
        if (self.type == SNKAttributeTypeInteger) {
            // Integers
            if ([self.value isKindOfClass:[NSNumber class]]) {
                dictionary[@"value"] = self.value;
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Integer values must be NSNumber"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeFloat) {
            // Floats
            if ([self.value isKindOfClass:[NSNumber class]]) {
                dictionary[@"value"] = self.value;
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Float values must be NSNumber"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeDecimal) {
            // Decimals
            if ([self.value isKindOfClass:[NSDecimalNumber class]]) {
                static NSNumberFormatter *numberFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    numberFormatter = [[NSNumberFormatter alloc] init];
                    [numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
                    [numberFormatter setGeneratesDecimalNumbers:YES];
                });
                dictionary[@"value"] = [numberFormatter stringFromNumber:(NSDecimalNumber *)self.value];
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Decimal values must be NSDecimalNumber"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeBoolean) {
            // Booleans
            if ([self.value isKindOfClass:[NSNumber class]]) {
                dictionary[@"value"] = self.value;
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Boolean values must be NSNumber"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeString) {
            // Strings
            if ([self.value isKindOfClass:[NSString class]]) {
                dictionary[@"value"] = self.value;
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"String values must be NSString"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeDate) {
            // Dates
            if ([self.value isKindOfClass:[NSDate class]]) {
                static NSDateFormatter *dateFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    dateFormatter = [[NSDateFormatter alloc] init];
                    [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
                    
                    NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
                    [dateFormatter setLocale:locale];
                });
                dictionary[@"value"] = [dateFormatter stringFromDate:(NSDate *)self.value];
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Date values must be NSDate"] userInfo:nil];
            }
        }
        else if (self.type == SNKAttributeTypeBinaryData) {
            // Binary datas
            if ([self.value isKindOfClass:[NSData class]]) {
                dictionary[@"value"] = [(NSData *)self.value snk_base64String];
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Binary data values must be NSData"] userInfo:nil];
            }
        }
    }
    else {
        dictionary[@"value"] = [NSNull null];
    }
    
    return dictionary;
}

@end


SNKAttributeType SNKAttributeTypeFromString(NSString *string) {
    if ([string isEqualToString:@"integer"]) {
        return SNKAttributeTypeInteger;
    }
    else if ([string isEqualToString:@"float"]) {
        return SNKAttributeTypeFloat;
    }
    else if ([string isEqualToString:@"decimal"]) {
        return SNKAttributeTypeDecimal;
    }
    else if ([string isEqualToString:@"boolean"]) {
        return SNKAttributeTypeBoolean;
    }
    else if ([string isEqualToString:@"string"]) {
        return SNKAttributeTypeString;
    }
    else if ([string isEqualToString:@"date"]) {
        return SNKAttributeTypeDate;
    }
    else if ([string isEqualToString:@"data"]) {
        return SNKAttributeTypeBinaryData;
    }
    return SNKAttributeTypeInteger;
}

NSString * SNKStringFromAttributeType(SNKAttributeType attributeType) {
    if (attributeType == SNKAttributeTypeInteger) {
        return @"integer";
    }
    else if (attributeType == SNKAttributeTypeFloat) {
        return @"float";
    }
    else if (attributeType == SNKAttributeTypeDecimal) {
        return @"decimal";
    }
    else if (attributeType == SNKAttributeTypeBoolean) {
        return @"boolean";
    }
    else if (attributeType == SNKAttributeTypeString) {
        return @"string";
    }
    else if (attributeType == SNKAttributeTypeDate) {
        return @"date";
    }
    else if (attributeType == SNKAttributeTypeBinaryData) {
        return @"data";
    }
    return @"integer";
}
