//
//  SNKRecord.m
//  SyncKit
//
//  Created by Logan Collins on 12/19/12.
//  Copyright (c) 2012 Sunflower Softworks. All rights reserved.
//

#import "SNKRecord.h"
#import "SNKRecord_Private.h"

#import "SNKProperty_Private.h"
#import "SNKAttribute.h"
#import "SNKRelationship.h"

#import "NSDictionary+SyncKitAdditions.h"


@implementation SNKRecord {
    NSMutableDictionary *_properties;
}

- (id)init {
    self = [super init];
    if (self) {
        _properties = [NSMutableDictionary dictionary];
    }
    return self;
}

- (id)initWithDictionary:(NSDictionary *)dictionary {
    self = [self init];
    if (self) {
        self.identifier = dictionary[@"identifier"];
        
        self.entityName = dictionary[@"entity"];
        
        id deletedValue = dictionary[@"is_deleted"];
        if ([deletedValue isKindOfClass:[NSString class]]) {
            self.deleted = [deletedValue isEqualToString:@"true"];
        }
        else if ([deletedValue isKindOfClass:[NSNumber class]]) {
            self.deleted = [deletedValue boolValue];
        }
        
        static NSDateFormatter *dateFormatter = nil;
        static dispatch_once_t onceToken;
        dispatch_once(&onceToken, ^{
            dateFormatter = [[NSDateFormatter alloc] init];
            [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
            
            NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
            [dateFormatter setLocale:locale];
        });
        
        id deletionDateValue = dictionary[@"deletion_date"];
        if ([deletionDateValue isKindOfClass:[NSString class]]) {
            self.deletionDate = [dateFormatter dateFromString:deletionDateValue];
        }
        
        // Attributes
        NSDictionary *attributeDicts = dictionary[@"attributes"];
        for (NSString *attributeName in attributeDicts) {
            NSDictionary *attributeDict = attributeDicts[attributeName];
            SNKAttribute *attribute = [[SNKAttribute alloc] initWithDictionary:attributeDict];
            [_properties setObject:attribute forKey:attributeName];
        }
        
        // Relationships
        NSDictionary *relationshipDicts = dictionary[@"relationships"];
        for (NSString *relationshipName in relationshipDicts) {
            NSDictionary *relationshipDict = relationshipDicts[relationshipName];
            SNKRelationship *relationship = [[SNKRelationship alloc] initWithDictionary:relationshipDict];
            [_properties setObject:relationship forKey:relationshipName];
        }
        
        id creationDateValue = dictionary[@"creation_date"];
        if ([creationDateValue isKindOfClass:[NSString class]]) {
            self.creationDate = [dateFormatter dateFromString:creationDateValue];
        }
        
        id modificationDateValue = dictionary[@"modification_date"];
        if ([modificationDateValue isKindOfClass:[NSString class]]) {
            self.modificationDate = [dateFormatter dateFromString:modificationDateValue];
        }
    }
    return self;
}

- (id)copyWithZone:(NSZone *)zone {
    SNKRecord *record = [[[self class] allocWithZone:zone] init];
    record.identifier = self.identifier;
    record.entityName = self.entityName;
    record.deleted = self.deleted;
    [record->_properties setDictionary:[[NSDictionary alloc] initWithDictionary:_properties copyItems:YES]];
    return record;
}

- (NSString *)description {
    return [NSString stringWithFormat:@"%@ { entity='%@', identifier='%@', deleted=%@ }", [super description], self.entityName, self.identifier, (self.deleted ? @"YES" : @"NO")];
}

- (NSDictionary *)dictionaryRepresentation {
    NSMutableDictionary *dictionary = [NSMutableDictionary dictionary];
    
    dictionary[@"identifier"] = self.identifier;
    dictionary[@"entity"] = self.entityName;
    dictionary[@"is_deleted"] = @(self.deleted);
    
    NSMutableDictionary *attributeDicts = [NSMutableDictionary dictionaryWithCapacity:[_properties count]];
    NSMutableDictionary *relationshipDicts = [NSMutableDictionary dictionaryWithCapacity:[_properties count]];
    
    for (NSString *name in _properties) {
        id property = _properties[name];
        NSDictionary *propertyDict = [property dictionaryRepresentation];
        
        if ([property isKindOfClass:[SNKAttribute class]]) {
            [attributeDicts setObject:propertyDict forKey:name];
        }
        else if ([property isKindOfClass:[SNKRelationship class]]) {
            [relationshipDicts setObject:propertyDict forKey:name];
        }
    }
    
    dictionary[@"attributes"] = attributeDicts;
    dictionary[@"relationships"] = relationshipDicts;
    
    static NSDateFormatter *dateFormatter = nil;
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        dateFormatter = [[NSDateFormatter alloc] init];
        [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
        
        NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
        [dateFormatter setLocale:locale];
    });
    
    if (self.deletionDate != nil) {
        dictionary[@"deletion_date"] = [dateFormatter stringFromDate:self.deletionDate];
    }
    else {
        dictionary[@"deletion_date"] = [NSNull null];
    }
    
    if (self.creationDate != nil) {
        dictionary[@"creation_date"] = [dateFormatter stringFromDate:self.creationDate];
    }
    else {
        dictionary[@"creation_date"] = [NSNull null];
    }
    
    if (self.modificationDate != nil) {
        dictionary[@"modification_date"] = [dateFormatter stringFromDate:self.modificationDate];
    }
    else {
        dictionary[@"modification_date"] = [NSNull null];
    }
    
    return dictionary;
}


#pragma mark -
#pragma mark Properties

- (NSDictionary *)propertiesByName {
    return [_properties copy];
}

- (NSArray *)properties {
    return [_properties allValues];
}

- (void)setProperties:(NSArray *)properties {
    [_properties removeAllObjects];
    for (SNKProperty *property in properties) {
        NSString *name = property.name;
        _properties[name] = property;
    }
}

- (SNKProperty *)propertyWithName:(NSString *)name {
    NSParameterAssert(name != nil);
    return _properties[name];
}

- (void)addProperty:(SNKProperty *)property {
    NSParameterAssert(property != nil);
    NSString *name = property.name;
    _properties[name] = property;
}

- (void)addProperties:(NSArray *)properties {
    for (SNKProperty *property in properties) {
        NSString *name = property.name;
        _properties[name] = property;
    }
}

- (void)removePropertyWithName:(NSString *)name {
    NSParameterAssert(name != nil);
    [_properties removeObjectForKey:name];
}

- (NSDictionary *)attributesByName {
    return [_properties snk_filteredDictionaryUsingBlock:^BOOL(id key, id value, BOOL *stop) {
        return [value isKindOfClass:[SNKAttribute class]];
    }];
}

- (NSDictionary *)relationshipsByName {
    return [_properties snk_filteredDictionaryUsingBlock:^BOOL(id key, id value, BOOL *stop) {
        return [value isKindOfClass:[SNKRelationship class]];
    }];
}

@end
