//
//  SNKRequest.m
//  SyncKit
//
//  Created by Logan Collins on 12/21/12.
//  Copyright (c) 2012 Sunflower Softworks. All rights reserved.
//

#import "SNKRequest.h"
#import "SNKRequest_Private.h"

#import "SNKService_Private.h"
#import "SNKCredentials_Private.h"

#import "NSData+SyncKitAdditions.h"
#import "NSMutableURLRequest+SyncKitAdditions.h"


#define REQUEST_DEFAULT_TIMEOUT 60.0


NSString * const SNKRequestDidFinishNotification = @"SNKRequestDidFinishNotification";


@interface SNKRequest () <NSURLConnectionDataDelegate>

@property (weak, readwrite) SNKService *syncService;
@property (copy, readwrite) NSURL *URL;
@property (readwrite) SNKRequestMethod requestMethod;
@property (readwrite) SNKRequestAuthorizationType authorizationType;
@property (strong, readwrite) SNKCredentials *credentials;
@property (copy, readwrite) NSDictionary *parameters;

@property (getter=isStarted) BOOL started;
@property (getter=isFinished) BOOL finished;
@property (getter=isCancelled) BOOL cancelled;

@end


@implementation SNKRequest {
    NSThread *_operationThread;
    NSURLConnection *_connection;
    void (^_completionHandler)(NSHTTPURLResponse *, NSData *, NSError *);
    NSHTTPURLResponse *_response;
    NSMutableData *_data;
    NSCondition *_condition;
}

+ (SNKRequest *)requestWithSyncService:(SNKService *)service URL:(NSURL *)URL requestMethod:(SNKRequestMethod)requestMethod authorizationType:(SNKRequestAuthorizationType)authorizationType credentials:(SNKCredentials *)credentials parameters:(NSDictionary *)parameters {
    SNKRequest *request = [[SNKRequest alloc] init];
    
    request.syncService = service;
    request.URL = URL;
    request.requestMethod = requestMethod;
    request.authorizationType = authorizationType;
    request.credentials = credentials;
    request.parameters = parameters;
    
    return request;
}

- (id)init {
    self = [super init];
    if (self) {
        
    }
    return self;
}

- (id)copyWithZone:(NSZone *)zone {
    SNKRequest *request = [[[self class] allocWithZone:zone] init];
    
    request.URL = self.URL;
    request.requestMethod = self.requestMethod;
    request.authorizationType = self.authorizationType;
    request.credentials = self.credentials;
    request.parameters = self.parameters;
    
    return request;
}

- (void)waitUntilFinished {
    [_condition lock];
    [_condition wait];
    [_condition unlock];
}


#pragma mark -
#pragma mark Request

- (NSURLRequest *)preparedURLRequest {
    NSMutableURLRequest *request = [[NSMutableURLRequest alloc] initWithURL:self.URL cachePolicy:NSURLRequestReloadIgnoringLocalCacheData timeoutInterval:REQUEST_DEFAULT_TIMEOUT];
    
    if (self.requestMethod == SNKRequestMethodGET) {
        [request setHTTPMethod:@"GET"];
        
        // Encode query parameters
        if (self.parameters != nil) {
            NSMutableString *query = [NSMutableString string];
            for (NSString *key in self.parameters) {
                NSString *value = self.parameters[key];
                if (![value isKindOfClass:[NSString class]]) {
                    value = [value description];
                }
                
                if ([query length] > 0) {
                    [query appendString:@"&"];
                }
                [query appendFormat:@"%@=%@", [key stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding], [value stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding]];
            }
            
            NSURL *newURL = [NSURL URLWithString:[[self.URL absoluteString] stringByAppendingFormat:@"?%@", query]];
            [request setURL:newURL];
        }
    }
    else if (self.requestMethod == SNKRequestMethodPOST) {
        [request setHTTPMethod:@"POST"];
        
        // Encode JSON parameters
        if (self.parameters != nil) {
            NSData *data = [NSJSONSerialization dataWithJSONObject:self.parameters options:0 error:nil];
            [request setValue:@"application/json" forHTTPHeaderField:@"Accept"];
            [request setValue:@"application/json" forHTTPHeaderField:@"Content-Type"];
            [request setValue:[NSString stringWithFormat:@"%d", (unsigned int)[data length]] forHTTPHeaderField:@"Content-Length"];
            [request setHTTPBody:data];
        }
    }
    else if (self.requestMethod == SNKRequestMethodDELETE) {
        [request setHTTPMethod:@"DELETE"];
    }
    
    // Authorization
    if (self.authorizationType == SNKRequestAuthorizationTypeOAuth) {
        [request snk_signWithOAuthConsumerKey:self.syncService.token consumerSecret:self.syncService.tokenSecret credentials:self.credentials];
    }
    
    // Allow gzip support for faster transfers
    [request setValue:@"gzip" forHTTPHeaderField:@"Accept-Encoding"];
    
    return request;
}

- (void)scheduleWithCompletionHandler:(void (^)(NSHTTPURLResponse *response, NSData *responseData, NSError *error))handler {
    if (_connection != nil) {
        @throw [NSException exceptionWithName:NSInternalInconsistencyException reason:@"Request has already been started" userInfo:nil];
    }
    
    _condition = [[NSCondition alloc] init];
    _completionHandler = [handler copy];
    
    _operationThread = [[NSThread alloc] initWithTarget:self selector:@selector(run) object:nil];
    [_operationThread start];
    
    self.started = YES;
}

- (void)run {
    NSURLRequest *request = [self preparedURLRequest];
    _connection = [[NSURLConnection alloc] initWithRequest:request delegate:self startImmediately:NO];
    [_connection scheduleInRunLoop:[NSRunLoop currentRunLoop] forMode:NSRunLoopCommonModes];
    [_connection start];
    
    // Spin the run loop
    while (!self.finished && !self.cancelled) {
        [[NSRunLoop currentRunLoop] runUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];
    }
}

- (void)cancel {
    [_connection cancel];
    
    self.cancelled = YES;
    self.finished = YES;
    
    [_condition lock];
    [_condition broadcast];
    [_condition unlock];
    
    [[NSNotificationCenter defaultCenter] postNotificationName:SNKRequestDidFinishNotification object:self];
}


#pragma mark -
#pragma mark Connection delegate

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response {
    _response = (NSHTTPURLResponse *)response;
    _data = [NSMutableData data];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data {
    [_data appendData:data];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
    self.finished = YES;
    
    if (_completionHandler != nil) {
        _completionHandler(_response, _data, nil);
    }
    
    [_condition lock];
    [_condition broadcast];
    [_condition unlock];
    
    [[NSNotificationCenter defaultCenter] postNotificationName:SNKRequestDidFinishNotification object:self];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error {
    self.finished = YES;
    
    if (_completionHandler != nil) {
        _completionHandler(_response, nil, error);
    }
    
    [_condition lock];
    [_condition broadcast];
    [_condition unlock];
    
    [[NSNotificationCenter defaultCenter] postNotificationName:SNKRequestDidFinishNotification object:self];
}

@end
