//
//  SNKService.h
//  SyncKit
//
//  Created by Logan Collins on 12/19/12.
//  Copyright (c) 2012 Sunflower Softworks. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <SyncKit/SyncKitDefines.h>
#import <SyncKit/SNKSyncSession.h>
#import <SyncKit/SNKRequest.h>


@class SNKAccount, SNKAuthorizationRequest, SNKCredentials;


/*!
 * @class SNKService
 * @abstract Manages communication between a client and sync server
 */
@interface SNKService : NSObject

/*!
 * @method initWithAPIEndpoint:
 * @abstract Creates a new service
 * 
 * @param endpoint
 * The API endpoint
 * 
 * @param token
 * The authorization token
 * 
 * @param secret
 * The authorization secret
 * 
 * @result An SNKService object
 */
- (id)initWithAPIEndpoint:(NSString *)endpoint token:(NSString *)token tokenSecret:(NSString *)secret credentials:(SNKCredentials *)credentials;


/*!
 * @property endpoint
 * @abstract The API endpoint
 * 
 * @result An NSString object
 */
@property (copy, readonly) NSString *endpoint;

/*!
 * @property token
 * @abstract The API token
 *
 * @result An NSString object
 */
@property (copy, readonly) NSString *token;

/*!
 * @property tokenSecret
 * @abstract The API token secret
 *
 * @result An NSString object
 */
@property (copy, readonly) NSString *tokenSecret;

/*!
 * @property credentials
 * @abstract The authorization credentials
 *
 * @result An SNKCredentials object
 */
@property (strong) SNKCredentials *credentials;


/*!
 * @property authorized
 * @abstract Whether the service is authorized (has a valid token)
 *
 * @result A BOOL value
 */
@property (readonly, getter=isAuthorized) BOOL authorized;

/*!
 * @method authorizeWithRequest:completionHandler:
 * @abstract Requests authorization from the server
 * 
 * @param authRequest
 * The authorization request to send
 * 
 * @param handler
 * The block to invoke upon request completion
 *
 * @discussion
 * If the request is successful, the new credentials are automatically applied to the receiver
 * and a SNKServiceAuthorizationDidChangeNotification notification is dispatched.
 *
 * @result An object conforming to the SNKRequest protocol
 */
- (id <SNKRequest>)authorizeWithRequest:(SNKAuthorizationRequest *)authRequest completionHandler:(void (^)(SNKAccount *account, NSError *error))handler;

/*!
 * @method deauthorizeWithCompletionHandler
 * @abstract Deauthorizes the current client for the user
 *
 * @param handler
 * The block to invoke upon request completion
 * 
 * @discussion
 * The current credentials will be invalidated by the server.
 *
 * If the request is successful, a SNKServiceAuthorizationDidChangeNotification notification is dispatched.
 * 
 * @result An object conforming to the SNKRequest protocol
 */
- (id <SNKRequest>)deauthorizeWithCompletionHandler:(void (^)(NSError *error))handler;

/*!
 * @method deauthorizeClientWithIdentifier:completionHandler
 * @abstract Deauthorizes a specified client for the user
 * 
 * @param identifier
 * The client identifier to deauthorize
 *
 * @param handler
 * The block to invoke upon request completion
 *
 * @discussion
 * The client's credentials will be invalidated by the server.
 *
 * @result An object conforming to the SNKRequest protocol
 */
- (id <SNKRequest>)deauthorizeClientWithIdentifier:(NSString *)identifier completionHandler:(void (^)(NSError *error))handler;

/*!
 * @method fetchAccountInfoWithCompletionHandler:
 * @abstract Fetches the user's account information
 * 
 * @param handler
 * The block to invoke upon request completion
 *
 * @result An object conforming to the SNKRequest protocol
 */
- (id <SNKRequest>)fetchAccountInfoWithCompletionHandler:(void (^)(SNKAccount *account, NSError *error))handler;

/*!
 * @method requestCredentialsResetWithEmail:completionHandler:
 * @abstract Requests a user be sent a credentials reset email
 *
 * @param email
 * The user's account email
 *
 * @param handler
 * The block to invoke upon request completion
 *
 * @result An object conforming to the SNKRequest protocol
 */
- (id <SNKRequest>)requestCredentialsResetWithEmail:(NSString *)email completionHandler:(void (^)(NSError *error))handler;


/*!
 * @method beginSyncSessionForSchema:delegate:options:
 * @abstract Begins a new sync session
 * 
 * @param schema
 * The schema name to sync
 * 
 * @param delegate
 * The delegate for the session
 * 
 * @param options
 * The options to use for the session
 * 
 * @discussion
 * If this method is called while another sync session is in progress for the same schema,
 * the new session will be queued until the previous finishes.
 * 
 * @result An SNKSyncSession object
 */
- (SNKSyncSession *)beginSyncSessionForSchema:(NSString *)schema delegate:(id <SNKSyncSessionDelegate>)delegate options:(NSDictionary *)options;

@end


/*!
 * @constant SNKServiceAuthorizationDidChangeNotification
 * @abstract Posted when a service's authorization status changes
 * 
 * @discussion
 * The notification object is the sync service for which authorization changed
 */
SYNCKIT_EXTERN NSString * const SNKServiceAuthorizationDidChangeNotification;

