//
//  SNKService.m
//  SyncKit
//
//  Created by Logan Collins on 12/19/12.
//  Copyright (c) 2012 Sunflower Softworks. All rights reserved.
//

#import "SNKService.h"
#import "SNKService_Private.h"

#import "SNKRequest_Private.h"
#import "SNKRecord_Private.h"
#import "SNKAccount_Private.h"
#import "SNKClient_Private.h"
#import "SNKAuthorizationRequest.h"
#import "SNKCredentials_Private.h"
#import "SNKSyncSession_Private.h"

#import "SyncKitErrors.h"

#import "NSData+SyncKitAdditions.h"


NSString * const SNKServiceAuthorizationDidChangeNotification = @"SNKServiceAuthorizationDidChangeNotification";


@implementation SNKService {
    NSOperationQueue *_operationQueue;
    NSMutableSet *_requests;
    NSMutableArray *_syncSessions;
}

- (id)init {
    @throw [NSException exceptionWithName:NSInvalidArgumentException reason:@"SNKService cannot be instantiated with -init." userInfo:nil];
}

- (id)initWithAPIEndpoint:(NSString *)endpoint token:(NSString *)token tokenSecret:(NSString *)secret credentials:(SNKCredentials *)credentials {
    NSParameterAssert(endpoint != nil);
    NSParameterAssert(token != nil);
    NSParameterAssert(secret != nil);
    
    self = [super init];
    if (self) {
        if ([endpoint hasSuffix:@"/"]) {
            endpoint = [endpoint substringWithRange:NSMakeRange(0, [endpoint length] - 1)];
        }
        self.endpoint = endpoint;
        self.token = token;
        self.tokenSecret = secret;
        self.credentials = credentials;
        
        _operationQueue = [[NSOperationQueue alloc] init];
        _operationQueue.name = @"com.sunflowersw.synckit.service.private";
        _operationQueue.maxConcurrentOperationCount = 1;
        
        _requests = [[NSMutableSet alloc] init];
        
        _syncSessions = [[NSMutableArray alloc] init];
        
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(requestDidFinish:) name:SNKRequestDidFinishNotification object:nil];
    }
    return self;
}

- (void)dealloc {
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}


#pragma mark -
#pragma mark Accessors

+ (NSSet *)keyPathsForValuesAffectingAuthorized {
    return [NSSet setWithObjects:@"token", @"tokenSecret", @"credentials", nil];
}

- (BOOL)isAuthorized {
    return (self.token != nil && self.tokenSecret != nil && self.credentials != nil);
}


#pragma mark -
#pragma mark Operations

- (void)performBlock:(void (^)(void))block {
    [_operationQueue addOperationWithBlock:block];
}

- (void)scheduleRequest:(SNKRequest *)request completionHandler:(void (^)(NSHTTPURLResponse *, NSData *, NSError *))handler {
    [self performBlock:^{
        [_requests addObject:request];
        [request scheduleWithCompletionHandler:handler];
    }];
}

- (void)requestDidFinish:(NSNotification *)notification {
    SNKRequest *request = [notification object];
    [self performBlock:^{
        [_requests removeObject:request];
    }];
}

- (void)scheduleSyncSession:(SNKSyncSession *)session {
    [self performBlock:^{
        [_syncSessions addObject:session];
        
        if ([_syncSessions count] == 1) {
            [session start];
        }
    }];
}

- (void)syncSessionDidFinish:(SNKSyncSession *)session {
    [self performBlock:^{
        [_syncSessions removeObject:session];
        
        if ([_syncSessions count] > 0) {
            SNKSyncSession *startSession = [_syncSessions objectAtIndex:0];
            [startSession start];
        }
    }];
}


#pragma mark -
#pragma mark Errors

- (NSError *)errorWithResponse:(NSHTTPURLResponse *)response responseData:(NSData *)responseData error:(NSError *)error {
    NSError *responseError = nil;
    
    if (response != nil) {
        // Construct an error from the response
        NSDictionary *errorInfo = nil;
        
        if (responseData != nil) {
            // Construct an error using the API error description
            NSError *jsonError = nil;
            NSDictionary *dictionary = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&jsonError];
            if ([dictionary isKindOfClass:[NSDictionary class]]) {
                NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
                
                NSString *description = dictionary[@"error"];
                if (description != nil) {
                    userInfo[NSLocalizedDescriptionKey] = description;
                }
                else {
                    if (error.userInfo[NSLocalizedDescriptionKey] != nil) {
                        userInfo[NSLocalizedDescriptionKey] = error.userInfo[NSLocalizedDescriptionKey];
                    }
                }
                
                NSString *reason = dictionary[@"message"];
                if (reason != nil) {
                    userInfo[NSLocalizedRecoverySuggestionErrorKey] = reason;
                }
                else {
                    if (error.userInfo[NSLocalizedFailureReasonErrorKey] != nil) {
                        userInfo[NSLocalizedFailureReasonErrorKey] = error.userInfo[NSLocalizedFailureReasonErrorKey];
                    }
                    if (error.userInfo[NSLocalizedRecoverySuggestionErrorKey] != nil) {
                        userInfo[NSLocalizedRecoverySuggestionErrorKey] = error.userInfo[NSLocalizedRecoverySuggestionErrorKey];
                    }
                }
                
                if (error != nil) {
                    userInfo[NSUnderlyingErrorKey] = error;
                }
                
                errorInfo = userInfo;
            }
            else {
                errorInfo = [jsonError userInfo];
            }
        }
        
        if (errorInfo == nil) {
            errorInfo = [error userInfo];
        }
        
        responseError = [NSError errorWithDomain:SNKSyncKitErrorDomain code:response.statusCode userInfo:errorInfo];
    }
    else {
        responseError = error;
    }
    
    return responseError;
}

- (NSError *)errorWithCode:(NSUInteger)errorCode {
    NSString *description = nil;
    NSString *failureReason = nil;
    
    if (errorCode == SNKInvalidRequestError) {
        
    }
    else if (errorCode == SNKNotAuthorizedError) {
        
    }
    
    NSMutableDictionary *userInfo = [NSMutableDictionary dictionary];
    if (description != nil) {
        userInfo[NSLocalizedDescriptionKey] = description;
    }
    if (failureReason != nil) {
        userInfo[NSLocalizedFailureReasonErrorKey] = failureReason;
    }
    
    return [NSError errorWithDomain:SNKSyncKitErrorDomain code:errorCode userInfo:userInfo];
}


#pragma mark -
#pragma mark Authorization

- (id <SNKRequest>)authorizeWithRequest:(SNKAuthorizationRequest *)authRequest completionHandler:(void (^)(SNKAccount *account, NSError *error))handler {
    NSParameterAssert(authRequest != nil);
    
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/users/authorize/", self.endpoint]];
    
    NSMutableDictionary *parameters = [NSMutableDictionary dictionary];
    parameters[@"email"] = authRequest.email;
    parameters[@"password"] = authRequest.password;
    parameters[@"create"] = (authRequest.shouldCreateAccount ? @"true" : @"false");
    parameters[@"identifier"] = authRequest.clientIdentifier;
    parameters[@"client"] = SNKLocalClientDictionary();
    
    SNKRequest *request = [SNKRequest requestWithSyncService:self URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeOAuth credentials:nil parameters:parameters];
    [self scheduleRequest:request completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        if (response.statusCode == 200) {
            NSError *localError = nil;
            NSDictionary *dictionary = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&localError];
            if ([dictionary isKindOfClass:[NSDictionary class]]) {
                SNKAccount *account = [[SNKAccount alloc] initWithDictionary:dictionary];
                
                NSString *token = dictionary[@"token"];
                NSString *secret = dictionary[@"secret"];
                
                SNKCredentials *credentials = [[SNKCredentials alloc] initWithOAuthToken:token tokenSecret:secret];
                self.credentials = credentials;
                
                if (handler != nil) {
                    handler(account, nil);
                }
                
                [[NSOperationQueue mainQueue] addOperationWithBlock:^{
                    [[NSNotificationCenter defaultCenter] postNotificationName:SNKServiceAuthorizationDidChangeNotification object:self];
                }];
            }
            else {
                if (handler != nil) {
                    handler(nil, [self errorWithCode:SNKInvalidResponseError]);
                }
            }
        }
        else {
            if (handler != nil) {
                handler(nil, [self errorWithResponse:response responseData:responseData error:error]);
            }
        }
    }];
    return request;
}

- (id <SNKRequest>)requestCredentialsResetWithEmail:(NSString *)email completionHandler:(void (^)(NSError *error))handler {
    NSParameterAssert(email != nil);
    
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/users/reset_password/", self.endpoint]];
    
    NSMutableDictionary *parameters = [NSMutableDictionary dictionary];
    parameters[@"email"] = email;
    
    SNKRequest *request = [SNKRequest requestWithSyncService:self URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeNone credentials:nil parameters:parameters];
    [self scheduleRequest:request completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        if (response.statusCode == 200) {
            if (handler != nil) {
                handler(nil);
            }
        }
        else {
            if (handler != nil) {
                handler([self errorWithResponse:response responseData:responseData error:error]);
            }
        }
    }];
    return request;
}

- (id <SNKRequest>)deauthorizeClientWithIdentifier:(NSString *)identifier completionHandler:(void (^)(NSError *error))handler; {
    NSParameterAssert(identifier != nil);
    
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/users/deauthorize/", self.endpoint]];
    
    NSMutableDictionary *parameters = [NSMutableDictionary dictionary];
    parameters[@"identifier"] = identifier;
    
    SNKRequest *request = [SNKRequest requestWithSyncService:self URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeOAuth credentials:self.credentials parameters:parameters];
    [self scheduleRequest:request completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        if (response.statusCode == 200) {
            NSError *localError = nil;
            NSDictionary *dictionary = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&localError];
            if ([dictionary isKindOfClass:[NSDictionary class]]) {
                if (handler != nil) {
                    handler(nil);
                }
                
                if ([identifier isEqualToString:self.credentials.OAuthToken]) {
                    [self deauthorize];
                }
            }
            else {
                if (handler != nil) {
                    handler([self errorWithCode:SNKInvalidResponseError]);
                }
            }
        }
        else {
            if (handler != nil) {
                handler([self errorWithResponse:response responseData:responseData error:error]);
            }
        }
    }];
    return request;
}

- (id <SNKRequest>)deauthorizeWithCompletionHandler:(void (^)(NSError *error))handler {
    if (self.credentials.OAuthToken != nil) {
        return [self deauthorizeClientWithIdentifier:self.credentials.OAuthToken completionHandler:handler];
    }
    else {
        if (handler != nil) {
            handler(nil);
        }
    }
    return nil;
}

- (void)deauthorize {
    self.credentials = nil;
    
    [[NSOperationQueue mainQueue] addOperationWithBlock:^{
        [[NSNotificationCenter defaultCenter] postNotificationName:SNKServiceAuthorizationDidChangeNotification object:self];
    }];
}

- (id <SNKRequest>)fetchAccountInfoWithCompletionHandler:(void (^)(SNKAccount *account, NSError *error))handler {
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/users/account/", self.endpoint]];
    
    SNKRequest *request = [SNKRequest requestWithSyncService:self URL:requestURL requestMethod:SNKRequestMethodGET authorizationType:SNKRequestAuthorizationTypeOAuth credentials:self.credentials parameters:nil];
    [self scheduleRequest:request completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        if (response.statusCode == 200 && responseData != nil) {
            NSError *localError = nil;
            NSDictionary *dictionary = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&localError];
            if ([dictionary isKindOfClass:[NSDictionary class]]) {
                SNKAccount *account = [[SNKAccount alloc] initWithDictionary:dictionary];
                if (handler != nil) {
                    handler(account, nil);
                }
            }
            else {
                if (handler != nil) {
                    handler(nil, [self errorWithCode:SNKInvalidResponseError]);
                }
            }
        }
        else {
            if (handler != nil) {
                handler(nil, [self errorWithResponse:response responseData:responseData error:error]);
            }
        }
    }];
    return request;
}


#pragma mark -
#pragma mark Sessions

- (SNKSyncSession *)beginSyncSessionForSchema:(NSString *)schema delegate:(id <SNKSyncSessionDelegate>)delegate options:(NSDictionary *)options {
    NSParameterAssert(delegate != nil);
    
    SNKSyncSession *session = [[SNKSyncSession alloc] initWithSyncService:self schema:schema delegate:delegate options:options];
    [self scheduleSyncSession:session];
    return session;
}

@end
