//
//  SNKSyncSession.h
//  SyncKit
//
//  Created by Logan Collins on 1/15/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <SyncKit/SyncKitDefines.h>


@class SNKService;
@protocol SNKSyncSessionDelegate;


/*!
 * @class SNKSyncSession
 * @abstract A sync session
 */
@interface SNKSyncSession : NSObject

/*!
 * @property syncService
 * @abstract The sync service managing the sync session
 * 
 * @result An SNKService object
 */
@property (strong, readonly) SNKService *syncService;

/*!
 * @property schema
 * @abstract The identifier of the schema for the session
 *
 * @result An NSString object
 */
@property (copy, readonly) NSString *schema;


/*!
 * @property delegate
 * @abstract The sync session delegate
 * 
 * @result An object conforming to the SNKSyncSessionDelegate protocol
 */
@property (unsafe_unretained, readonly) id <SNKSyncSessionDelegate> delegate;


/*!
 * @property syncing
 * @abstract Whether the coordinator is currently performing a sync
 *
 * @result A BOOL value
 */
@property (readonly, getter=isSyncing) BOOL syncing;

/*!
 * @method cancel
 * @abstract Cancels the sync session
 * 
 * @discussion
 * No further delegate methods will be invoked.
 */
- (void)cancel;


/*!
 * @property numberOfRecordsPushed
 * @abstract The number of records pushed during the session
 * 
 * @result An NSUInteger value
 */
@property (readonly) NSUInteger numberOfRecordsPushed;

/*!
 * @property numberOfRecordsPulled
 * @abstract The number of records pulled during the session
 *
 * @result An NSUInteger value
 */
@property (readonly) NSUInteger numberOfRecordsPulled;

@end


/*!
 * @protocol SNKSyncSessionDelegate
 * @abstract Implemented by a sync session delegate
 */
@protocol SNKSyncSessionDelegate <NSObject>

@optional

/*!
 * @method syncSessionWillStart:
 * @abstract Invoked when the sync session is about to start
 *
 * @param session
 * The sync session
 */
- (void)syncSessionWillStart:(SNKSyncSession *)session;

/*!
 * @method syncSessionDidFinish:
 * @abstract Invoked when the sync session has finished
 *
 * @param session
 * The sync session
 */
- (void)syncSessionDidFinish:(SNKSyncSession *)session;

/*!
 * @method syncSession:didFailWithError:
 * @abstract Invoked when the sync session has failed
 *
 * @param session
 * The sync session
 *
 * @param error
 * The error generated by the failure
 */
- (void)syncSession:(SNKSyncSession *)session didFailWithError:(NSError *)error;


@required

/*!
 * @method syncSessionEntitiesToSync:
 * @abstract Requests the names of entities that should be synced
 *
 * @result An NSArray of NSString objects
 */
- (NSArray *)syncSessionEntitiesToSync:(SNKSyncSession *)session;

/*!
 * @method syncSession:recordsForEntity:moreComing:error:
 * @abstract Requests all records for an entity
 *
 * @param session
 * The sync session
 *
 * @param entityName
 * The entity name for which records are requested
 *
 * @param moreComing
 * If set to YES, the coordinator will continue asking for records
 *
 * @param outError
 * Set to an error object in the case of a failure
 *
 * @discussion
 * This method is called when an entity is slow-synced (when all records are requested.)
 *
 * The client may set `moreComing` to YES to allow the sync session
 * to request records in batches. Once this method returns with `moreComing`
 * set to NO (the default) the sync session will not ask for more records.
 *
 * If an error is encountered this method should return nil and set the error to `outError`.
 * In this case, the sync session will automatically fail.
 *
 * @result An NSDictionary object (NSString -> SNKRecord)
 */
- (NSDictionary *)syncSession:(SNKSyncSession *)session recordsForEntity:(NSString *)entityName moreComing:(BOOL *)moreComing error:(NSError **)outError;

/*!
 * @method syncSession:applyRecords:moreComing:error:
 * @abstract Allows the delegate to apply a set of pulled records
 *
 * @param session
 * The sync session
 *
 * @param records
 * The dictionary of records to apply (NSString -> SNKRecord)
 * 
 * @param moreComing
 * If YES, the coordinator will deliver more records after this callback
 *
 * @param outError
 * Set to an error object in the case of a failure
 *
 * @discussion
 * This method may be called multiple times for a sync session if the server
 * delivers pulled records in batches. In this case, care should be taken to
 * maintain inconsistent state between records (in cases where a to-many or
 * to-one relationship's counterpart may not have arrived yet).
 *
 * If an error is encountered, this method should return NO and set the error to `outError`.
 * In this case, the sync session will automatically fail.
 *
 * @result A BOOL value
 */
- (BOOL)syncSession:(SNKSyncSession *)session applyRecords:(NSDictionary *)records moreComing:(BOOL)moreComing error:(NSError **)outError;


@optional

/*!
 * @method syncSession:shouldPerformFastSyncForEntity:
 * @abstract Whether the sync session should perform a fast-sync for an entity
 *
 * @param session
 * The sync session
 *
 * @param entityName
 * The entity name
 *
 * @discussion
 * By returning NO, the delegate can indicate that it either has never performed
 * a sync of the entity or otherwise needs to push all records for the entity.
 *
 * If the delegate returns YES for an entity, the session may still initiate a
 * slow-sync if the server requests it.
 *
 * @result An NSString object, or nil
 */
- (BOOL)syncSession:(SNKSyncSession *)session shouldPerformFastSyncForEntity:(NSString *)entityName;

/*!
 * @method syncSession:recordsChangedForEntity:moreComing:error:
 * @abstract Requests records that have changed for an entity
 *
 * @param session
 * The sync session
 * 
 * @param entityName
 * The entity name for which records are requested
 *
 * @param moreComing
 * If set to YES, the coordinator will continue asking for records
 *
 * @param outError
 * Set to an error object in the case of a failure
 * 
 * @discussion
 * This method is called when an entity is fast-synced (when only changed records are requested.)
 *
 * The client may set `moreComing` to YES to allow the sync session
 * to request records in batches. Once this method returns with `moreComing`
 * set to NO (the default) the sync session will not ask for more records.
 * 
 * If an error is encountered this method should return nil and set the error to `outError`.
 * In this case, the sync session will automatically fail.
 *
 * @result An NSDictionary object (NSString -> SNKRecord)
 */
- (NSDictionary *)syncSession:(SNKSyncSession *)session recordsChangedForEntity:(NSString *)entityName moreComing:(BOOL *)moreComing error:(NSError **)outError;

/*!
 * @method syncSessionShouldFinalize:error:
 * @abstract Whether the sync session should finalize (marking the client as up-to-date)
 *
 * @param session
 * The sync session
 *
 * @param outError
 * Set to an error object in the case of a failure
 * 
 * @discussion
 * This method is called when all records have been delivered to the client, and gives
 * the client a chance to commit all changes and validate the state of its records.
 *
 * If an error is encountered during this stage, the client can return NO, along with an error object,
 * to indicate that the sync session should not be finalized.
 */
- (BOOL)syncSessionShouldFinalize:(SNKSyncSession *)session error:(NSError *__autoreleasing *)outError;


/*!
 * @method syncSession:didEncounterConflicts:responseHandler:
 * @abstract Allows the delegate to handle conflicts encountered while performing a sync operation
 * 
 * @param session
 * The sync session
 * 
 * @param conflicts
 * The array of conflicts that need resolution
 * 
 * @param resolutionHandler
 * The block that must be invoked by the delegate after resolution of all conflicts
 * 
 * @discussion
 * The resolution block must be called for the sync session to continue, or else the thread coordinating the sync will spin indefinitely.
 * All conflict objects should be resolved by setting their resolutionAction property before calling the resolution handler.
 */
- (void)syncSession:(SNKSyncSession *)session didEncounterConflicts:(NSArray *)conflicts resolutionHandler:(void (^)(void))resolutionHandler;

@end


/*!
 * @constant SNKSyncSessionSlowSyncOption
 * @abstract Requests a slow sync (in which the client pushes, and server delivers, all records)
 *
 * @discussion
 * The value of this option should be an NSNumber -> BOOL.
 */
SYNCKIT_EXTERN NSString * const SNKSyncSessionSlowSyncOption;

