//
//  SNKSyncSession.m
//  SyncKit
//
//  Created by Logan Collins on 1/15/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "SNKSyncSession.h"
#import "SNKSyncSession_Private.h"

#import "SNKService_Private.h"
#import "SNKRecord_Private.h"
#import "SNKRequest_Private.h"
#import "SNKClient_Private.h"

#import "SyncKitErrors.h"


NSString * const SNKSyncSessionSlowSyncOption = @"SNKSyncSessionSlowSyncOption";


@implementation SNKSyncSession {
    SNKRequest *_currentRequest;
    SNKCredentials *_credentials;
    NSArray *_entitiesToSync;
    NSMutableDictionary *_entityOptions;
    NSString *_sessionIdentifier;
}

- (id)initWithSyncService:(SNKService *)service schema:(NSString *)schema delegate:(id <SNKSyncSessionDelegate>)delegate options:(NSDictionary *)options {
    NSParameterAssert([schema length] > 0);
    
    self = [super init];
    if (self) {
        self.syncService = service;
        self.schema = schema;
        self.delegate = delegate;
        self.options = options;
    }
    return self;
}

- (void)start {
    if (self.syncing) {
        @throw [NSException exceptionWithName:NSInternalInconsistencyException reason:@"Sync session has already started" userInfo:nil];
    }
    
    self.syncing = YES;
    
    _credentials = self.syncService.credentials;
    
    if ([self.delegate respondsToSelector:@selector(syncSessionWillStart:)]) {
        [self.delegate syncSessionWillStart:self];
    }
    
    NSArray *entitiesToSync = nil;
    if ([self.delegate respondsToSelector:@selector(syncSessionEntitiesToSync:)]) {
        entitiesToSync = [self.delegate syncSessionEntitiesToSync:self];
    }
    if (entitiesToSync == nil) {
        self.syncing = NO;
        if ([self.delegate respondsToSelector:@selector(syncSessionDidFinish:)]) {
            [self.delegate syncSessionDidFinish:self];
        }
        return;
    }
    _entitiesToSync = entitiesToSync;
    
    [self performPrepareSync];
}

- (void)performPrepareSync {
    // Phase 0: Prepare for sync
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/records/prepare_sync/", self.syncService.endpoint]];
    NSDictionary *requestParameters = @{ @"schema": self.schema, @"entities": _entitiesToSync };
    
    SNKRequest *prepareRequest = [SNKRequest requestWithSyncService:self.syncService URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeOAuth credentials:_credentials parameters:requestParameters];
    _currentRequest = prepareRequest;
    [self.syncService scheduleRequest:prepareRequest completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        NSError *localError = nil;
        NSDictionary *entitiesInfo = nil;
        
        if (response.statusCode == 200 && responseData != nil) {
            // Success
            NSError *jsonError = nil;
            id value = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&jsonError];
            if ([value isKindOfClass:[NSDictionary class]]) {
                entitiesInfo = value[@"entities"];
                _sessionIdentifier = value[@"session"];
            }
            else {
                localError = [self.syncService errorWithCode:SNKInvalidResponseError];
            }
        }
        else if (error != nil) {
            // Error
            localError = error;
        }
        else {
            localError = [self.syncService errorWithResponse:response responseData:responseData error:error];
        }
        
        NSMutableDictionary *entityOptionsByEntityName = [NSMutableDictionary dictionaryWithCapacity:[_entitiesToSync count]];
        for (NSString *entityName in _entitiesToSync) {
            NSMutableDictionary *entityOptions = entityOptionsByEntityName[entityName];
            if (entityOptions == nil) {
                entityOptions = [NSMutableDictionary dictionary];
                entityOptionsByEntityName[entityName] = entityOptions;
            }
            
            BOOL shouldPerformFastSync = NO;
            
            NSDate *entityLastSyncDate = nil;
            NSString *lastSyncDateString = entitiesInfo[entityName][@"last_sync_date"];
            if (lastSyncDateString != nil && ![lastSyncDateString isEqual:[NSNull null]]) {
                static NSDateFormatter *dateFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    dateFormatter = [[NSDateFormatter alloc] init];
                    [dateFormatter setDateFormat:@"eee, d MMM y H:mm:ss Z"];
                    
                    NSLocale *locale = [[NSLocale alloc] initWithLocaleIdentifier:@"en_US_POSIX"];
                    [dateFormatter setLocale:locale];
                });
                entityLastSyncDate = [dateFormatter dateFromString:lastSyncDateString];
            }
            
            if (entityLastSyncDate != nil) {
                if ([self.delegate respondsToSelector:@selector(syncSession:shouldPerformFastSyncForEntity:)]) {
                    shouldPerformFastSync = [self.delegate syncSession:self shouldPerformFastSyncForEntity:entityName];
                }
            }
            
            entityOptions[@"fast"] = @(shouldPerformFastSync);
        }
        _entityOptions = entityOptionsByEntityName;
        
        
        if (localError != nil) {
            if ([self.delegate respondsToSelector:@selector(syncSession:didFailWithError:)]) {
                [self.delegate syncSession:self didFailWithError:localError];
            }
            
            self.syncing = NO;
            [self.syncService syncSessionDidFinish:self];
            _currentRequest = nil;
        }
        else {
            [self performPush];
        }
    }];
}

- (void)performPush {
    // Phase I: Push
    NSError *error = nil;
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/records/sync/", self.syncService.endpoint]];
    NSMutableDictionary *parameters = [NSMutableDictionary dictionary];
    
    parameters[@"session"] = _sessionIdentifier;
    parameters[@"schema"] = self.schema;
    parameters[@"entities"] = _entitiesToSync;
    
    if ([self.options[SNKSyncSessionSlowSyncOption] boolValue]) {
        parameters[@"reset"] = @YES;
    }
    
    // Get records from delegate
    NSMutableDictionary *pushRecords = [NSMutableDictionary dictionary];
    for (NSString *entityName in _entitiesToSync) {
        NSDictionary *options = _entityOptions[entityName];
        BOOL fastSync = [options[@"fast"] boolValue];
        
        if (fastSync && [self.delegate respondsToSelector:@selector(syncSession:recordsChangedForEntity:moreComing:error:)]) {
            // Fast-sync
            BOOL moreComing = NO;
            do {
                NSDictionary *records = [self.delegate syncSession:self recordsChangedForEntity:entityName moreComing:&moreComing error:&error];
                
                if (records != nil) {
                    [pushRecords addEntriesFromDictionary:records];
                }
                else {
                    break;
                }
            }
            while (moreComing);
        }
        else {
            // Slow-sync
            BOOL moreComing = NO;
            do {
                NSDictionary *records = [self.delegate syncSession:self recordsForEntity:entityName moreComing:&moreComing error:&error];
                
                if (records != nil) {
                    [pushRecords addEntriesFromDictionary:records];
                }
                else {
                    break;
                }
            }
            while (moreComing);
        }
    }
    
    if (error == nil) {
        // Records
        NSMutableDictionary *formattedPushRecords = [NSMutableDictionary dictionaryWithCapacity:[pushRecords count]];
        for (id key in pushRecords) {
            NSString *newKey = nil;
            if ([key isKindOfClass:[NSString class]]) {
                newKey = key;
            }
            else {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Invalid record identifier: %@", key] userInfo:nil];
            }
            
            SNKRecord *record = pushRecords[key];
            NSDictionary *recordDict = [record dictionaryRepresentation];
            formattedPushRecords[newKey] = recordDict;
        }
        parameters[@"records"] = formattedPushRecords;
        
        // Client Info
        parameters[@"client"] = SNKLocalClientDictionary();
        
        // Build push request
        SNKRequest *pushRequest = [SNKRequest requestWithSyncService:self.syncService URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeOAuth credentials:_credentials parameters:parameters];
        _currentRequest = pushRequest;
        [self.syncService scheduleRequest:pushRequest completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *connectionError) {
            NSError *localError = nil;
            NSMutableDictionary *newRecords = nil;
            
            if (connectionError != nil) {
                // Error
                localError = connectionError;
            }
            else if (response.statusCode == 200 && responseData != nil) {
                // Success
                self.numberOfRecordsPushed = [formattedPushRecords count];
                
                NSError *jsonError = nil;
                id value = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&jsonError];
                if ([value isKindOfClass:[NSDictionary class]]) {
                    NSDictionary *result = (NSDictionary *)value;
                    
                    NSDictionary *recordDicts = result[@"records"];
                    newRecords = [NSMutableDictionary dictionaryWithCapacity:[recordDicts count]];
                    for (NSString *uuidString in recordDicts) {
                        NSDictionary *recordDict = recordDicts[uuidString];
                        SNKRecord *record = [[SNKRecord alloc] initWithDictionary:recordDict];
                        newRecords[uuidString] = record;
                    }
                    
                    BOOL moreComing = NO;
                    
                    if ([newRecords count] > 0 && [self.delegate respondsToSelector:@selector(syncSession:applyRecords:moreComing:error:)]) {
                        NSError *delegateError = nil;
                        if (![self.delegate syncSession:self applyRecords:newRecords moreComing:moreComing error:&delegateError]) {
                            localError = delegateError;
                        }
                    }
                }
                else {
                    localError = [self.syncService errorWithCode:SNKInvalidResponseError];
                }
            }
            else {
                localError = [self.syncService errorWithResponse:response responseData:responseData error:connectionError];
            }
            
            if (localError != nil) {
                if ([self.delegate respondsToSelector:@selector(syncSession:didFailWithError:)]) {
                    [self.delegate syncSession:self didFailWithError:localError];
                }
            }
            else {
                BOOL shouldFinalize = YES;
                if ([self.delegate respondsToSelector:@selector(syncSessionShouldFinalize:error:)]) {
                    shouldFinalize = [self.delegate syncSessionShouldFinalize:self error:&localError];
                }
                
                if (shouldFinalize) {
                    self.numberOfRecordsPulled = [newRecords count];
                    [self performFinalizeSync];
                }
                else {
                    if ([self.delegate respondsToSelector:@selector(syncSession:didFailWithError:)]) {
                        [self.delegate syncSession:self didFailWithError:localError];
                    }
                }
            }
            
            self.syncing = NO;
            [self.syncService syncSessionDidFinish:self];
            _currentRequest = nil;
        }];
    }
    else {
        if ([self.delegate respondsToSelector:@selector(syncSession:didFailWithError:)]) {
            [self.delegate syncSession:self didFailWithError:error];
        }
        self.syncing = NO;
        [self.syncService syncSessionDidFinish:self];
        _currentRequest = nil;
    }
}

- (void)performFinalizeSync {
    // Finalize sync
    NSURL *requestURL = [NSURL URLWithString:[NSString stringWithFormat:@"%@/records/finalize_sync/", self.syncService.endpoint]];
    NSDictionary *requestParameters = @{ @"session": _sessionIdentifier, @"schema": self.schema, @"entities": _entitiesToSync };
    
    SNKRequest *finalizeRequest = [SNKRequest requestWithSyncService:self.syncService URL:requestURL requestMethod:SNKRequestMethodPOST authorizationType:SNKRequestAuthorizationTypeOAuth credentials:_credentials parameters:requestParameters];
    _currentRequest = finalizeRequest;
    [self.syncService scheduleRequest:finalizeRequest completionHandler:^(NSHTTPURLResponse *response, NSData *responseData, NSError *error) {
        NSError *localError = nil;
//        NSDictionary *entitiesInfo = nil;
        
        if (response.statusCode == 200 && responseData != nil) {
            // Success
            NSError *jsonError = nil;
            id value = [NSJSONSerialization JSONObjectWithData:responseData options:0 error:&jsonError];
            if ([value isKindOfClass:[NSDictionary class]]) {
//                entitiesInfo = value[@"entities"];
            }
            else {
                localError = [self.syncService errorWithCode:SNKInvalidResponseError];
            }
        }
        else if (error != nil) {
            // Error
            localError = error;
        }
        else {
            localError = [self.syncService errorWithResponse:response responseData:responseData error:error];
        }
        
        if (localError != nil) {
            if ([self.delegate respondsToSelector:@selector(syncSession:didFailWithError:)]) {
                [self.delegate syncSession:self didFailWithError:localError];
            }
        }
        else {
            if ([self.delegate respondsToSelector:@selector(syncSessionDidFinish:)]) {
                [self.delegate syncSessionDidFinish:self];
            }
        }
        
        self.syncing = NO;
        [self.syncService syncSessionDidFinish:self];
        _currentRequest = nil;
    }];
}

- (void)cancel {
    [_currentRequest cancel];
    _currentRequest = nil;
    self.syncing = NO;
    [self.syncService syncSessionDidFinish:self];
}

@end
